# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast

import logging
from enum import IntEnum

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gtk
from nbxmpp.modules.vcard4 import VCard
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import AnnotationNote
from nbxmpp.structs import SoftwareVersionResult
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import ged
from gajim.common import types
from gajim.common.const import AvatarSize
from gajim.common.const import SimpleClientState
from gajim.common.events import SubscribedPresenceReceived
from gajim.common.events import UnsubscribedPresenceReceived
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.contacts import ResourceContact
from gajim.common.util.status import get_uf_show
from gajim.common.util.user_strings import get_uf_affiliation
from gajim.common.util.user_strings import get_uf_role

from gajim.gtk.alert import AlertDialog
from gajim.gtk.alert import CancelDialogResponse
from gajim.gtk.alert import ConfirmationAlertDialog
from gajim.gtk.alert import DialogResponse
from gajim.gtk.builder import get_builder
from gajim.gtk.contact_name_entry import ContactNameEntry
from gajim.gtk.contact_settings import ContactSettings
from gajim.gtk.crypto_trust_manager import CryptoTrustManager
from gajim.gtk.sidebar_switcher import SideBarSwitcher
from gajim.gtk.structs import AccountJidParam
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.vcard_grid import VCardGrid
from gajim.gtk.window import GajimAppWindow

log = logging.getLogger("gajim.gtk.contact_info")


ContactT = BareContact | GroupchatParticipant


class Column(IntEnum):
    IN_GROUP = 0
    GROUP_NAME = 1


class ContactInfo(GajimAppWindow, EventHelper):
    def __init__(
        self,
        account: str,
        contact: GroupchatParticipant | BareContact,
        page: str | None = None,
    ) -> None:
        GajimAppWindow.__init__(
            self,
            name="ContactInfo",
            title=_("Contact Information"),
            default_width=700,
            default_height=600,
        )

        EventHelper.__init__(self)

        # Window management needs both account and contact attributes set
        self.account = account
        self.contact = contact

        if isinstance(contact, GroupchatParticipant):
            self._is_in_roster = False
            self._bare_contact = contact.get_real_contact()
            if self._bare_contact is not None:
                self._is_in_roster = self._bare_contact.is_in_roster
        else:
            self._bare_contact = contact
            self._is_in_roster = contact.is_in_roster

        self._client = app.get_client(account)
        self._client.connect_signal("state-changed", self._on_client_state_changed)

        self._ui = get_builder("contact_info.ui")

        self._tasks: list[Task] = []
        self._devices: dict[str, DeviceInfo] = {}

        self._switcher = SideBarSwitcher()

        toolbar = Adw.ToolbarView(content=self._switcher)
        toolbar.add_top_bar(Adw.HeaderBar())

        self._sidebar_page = Adw.NavigationPage(
            title=self.contact.name, tag="sidebar", child=toolbar
        )

        toolbar = Adw.ToolbarView(content=self._ui.main_stack)
        toolbar.add_top_bar(Adw.HeaderBar())

        content_page = Adw.NavigationPage(title=" ", tag="content", child=toolbar)

        nav = Adw.NavigationSplitView(sidebar=self._sidebar_page, content=content_page)

        self.set_child(nav)

        self._switcher.set_with_stack(self._ui.main_stack, visible=False)

        self._load_avatar()

        self._fill_information_page()
        self._fill_settings_page()
        self._fill_encryption_page()
        self._fill_device_info()
        self._fill_groups_page()
        self._fill_note_page()

        if page is not None:
            self._switcher.activate_item(page)
        else:
            self._switcher.activate_item("information")

        self._connect(
            self._ui.tree_selection, "changed", self._on_group_selection_changed
        )
        self._connect(self._ui.toggle_renderer, "toggled", self._on_group_toggled)
        self._connect(self._ui.text_renderer, "edited", self._on_group_name_edited)

        self._connect(
            self._ui.from_subscription_switch,
            "state-set",
            self._on_from_subscription_switch_toggled,
        )
        self._connect(
            self._ui.to_subscription_button,
            "clicked",
            self._on_to_subscription_button_clicked,
        )
        self._connect(
            self._ui.group_add_button, "clicked", self._on_group_add_button_clicked
        )
        self._connect(
            self._ui.group_remove_button,
            "clicked",
            self._on_group_remove_button_clicked,
        )

        self.register_events(
            [
                (
                    "subscribed-presence-received",
                    ged.GUI1,
                    self._on_subscribed_presence_received,
                ),
                (
                    "unsubscribed-presence-received",
                    ged.GUI1,
                    self._on_unsubscribed_presence_received,
                ),
            ]
        )

    def _cleanup(self) -> None:
        for task in self._tasks:
            task.cancel()

        self._save_annotation()

        for device_info in self._devices.values():
            device_info.run_destroy()
            self._ui.devices_page.remove(device_info)

        self._devices.clear()

        self._switcher.run_destroy()
        del self._switcher

        self._vcard_grid.run_destroy()
        del self._vcard_grid

        self._client.disconnect_all_from_obj(self)
        self._disconnect_all()
        self.unregister_events()

    def _on_client_state_changed(
        self, _client: types.Client, _signal_name: str, state: SimpleClientState
    ) -> None:
        self._ui.subscription_listbox.set_sensitive(state.is_connected)

        if state.is_connected:
            self._ui.groups_page_stack.set_visible_child_name("groups")
            self._ui.notes_page_stack.set_visible_child_name("notes")
        else:
            self._ui.groups_page_stack.set_visible_child_name("offline")
            self._ui.notes_page_stack.set_visible_child_name("offline")

    def _on_contact_name_updated(self, _widget: ContactNameEntry, name: str) -> None:
        self._sidebar_page.set_title(name)

    def _fill_information_page(self) -> None:
        contact_name_entry = ContactNameEntry(
            contact=self.contact, editable=self._is_in_roster
        )
        self._connect(contact_name_entry, "name-updated", self._on_contact_name_updated)
        self._ui.contact_name_controls_box.append(contact_name_entry)

        self._vcard_grid = VCardGrid(self.account)
        self._ui.vcard_box.append(self._vcard_grid)
        if self._client.is_available() and self._bare_contact is not None:
            self._client.get_module("VCard4").request_vcard(
                jid=self._bare_contact.jid, callback=self._on_vcard_received
            )

        if self._bare_contact is not None:
            jid = str(self._bare_contact.get_address())
        else:
            jid = str(self.contact.get_address())

        self._ui.contact_jid_label.set_text(jid)
        self._ui.contact_jid_label.set_tooltip_text(jid)

        if isinstance(self.contact, GroupchatParticipant):
            self._ui.affiliation_info.set_subtitle(
                get_uf_affiliation(self.contact.affiliation)
            )
            self._ui.role_info.set_subtitle(get_uf_role(self.contact.role))
            self._ui.group_chat_info.set_visible(True)

        self._switcher.set_item_visible("information", True)

    def _fill_encryption_page(self) -> None:
        if self._bare_contact is None:
            return

        self._ui.encryption_omemo_box.set_child(
            CryptoTrustManager("OMEMO", self._bare_contact.account, self._bare_contact)
        )
        self._switcher.set_item_visible("encryption-omemo", True)

        self._ui.encryption_openpgp_box.set_child(
            CryptoTrustManager(
                "OpenPGP", self._bare_contact.account, self._bare_contact
            )
        )
        self._switcher.set_item_visible("encryption-openpgp", True)

    def _fill_note_page(self) -> None:
        if self._bare_contact is None or not self._is_in_roster:
            return

        note = self._client.get_module("Annotations").get_note(self._bare_contact.jid)
        if note is not None:
            self._ui.textview_annotation.get_buffer().set_text(note.data)

        server_disco = self._client.get_module("Discovery").server_info
        if server_disco is not None and server_disco.supports(Namespace.PRIVATE):
            # Hide the Notes page if private storage is not available, because
            # roster notes cannot be stored without.
            # Since there is no disco mechanism for private storage, we rely on
            # Delimiter as a "proxy" for the availability of private storage.
            self._switcher.set_item_visible("notes", True)

    def _fill_settings_page(self) -> None:
        if self._bare_contact is None or not self._is_in_roster:
            return

        self._ui.from_subscription_switch.set_active(self._bare_contact.is_subscribed)

        self._ui.subscription_listbox.set_sensitive(self._client.is_available())

        if self._bare_contact.subscription in ("to", "both"):
            self._ui.to_subscription_stack.set_visible_child_name("checkmark")

        elif self._bare_contact.ask == "subscribe":
            self._ui.to_subscription_stack.set_visible_child_name("request")
            self._ui.request_stack.set_visible_child_name("requested")
        else:
            self._ui.to_subscription_stack.set_visible_child_name("request")
            self._ui.request_stack.set_visible_child_name("cross")

        self._switcher.set_item_visible("settings", True)

        if isinstance(self.contact, BareContact):
            # All settings for PM are derived from the groupchat settings
            # so we dont need to show them
            contact_settings = ContactSettings(self.account, self._bare_contact.jid)
            self._ui.settings_box.insert_child_after(
                contact_settings, self._ui.subscription_box
            )

        params = AccountJidParam(account=self.account, jid=self.contact.jid)
        self._ui.remove_history_button.set_action_target_value(params.to_variant())
        self._ui.remove_history_button.set_action_name("app.remove-history")

    def _fill_groups_page(self) -> None:
        if self._bare_contact is None or not self._is_in_roster:
            return

        if not self._client.is_available():
            return

        model = self._ui.groups_treeview.get_model()
        assert isinstance(model, Gtk.ListStore)
        model.set_sort_column_id(Column.GROUP_NAME, Gtk.SortType.ASCENDING)
        groups = self._client.get_module("Roster").get_groups()
        for group in groups:
            is_in_group = group in self._bare_contact.groups
            model.append([is_in_group, group])

        self._switcher.set_item_visible("groups", True)

    def _fill_device_info(self) -> None:
        if self._bare_contact is None or not self._is_in_roster:
            return

        contacts = list(self._bare_contact.iter_resources())
        if not contacts:
            return

        for resource_contact in contacts:
            device_grid = DeviceInfo(resource_contact)
            self._devices[resource_contact.resource] = device_grid
            self._ui.devices_page.add(device_grid)
            self._query_device(resource_contact)

        self._ui.devices_stack.set_visible_child_name("devices")
        self._switcher.set_item_visible("devices", True)

    def _query_device(self, contact: ResourceContact) -> None:
        software_module = self._client.get_module("SoftwareVersion")
        task = software_module.request_software_version(
            contact.jid, callback=self._set_os_info, user_data=contact.resource
        )
        self._tasks.append(task)

        task = self._client.get_module("EntityTime").request_entity_time(
            contact.jid, callback=self._set_entity_time, user_data=contact.resource
        )
        self._tasks.append(task)

    def _on_vcard_received(self, jid: JID, vcard: VCard | None) -> None:
        if vcard is None:
            vcard = VCard()
        self._vcard_grid.set_vcard(vcard)
        self._ui.vcard_box.set_visible(self._vcard_grid.get_row_count() != 0)

    def _load_avatar(self) -> None:
        scale = self.get_scale_factor()
        texture = self.contact.get_avatar(AvatarSize.VCARD, scale, add_show=False)
        self._ui.avatar_image.set_pixel_size(AvatarSize.VCARD)
        self._ui.avatar_image.set_from_paintable(texture)

    def _set_os_info(self, task: Task) -> None:
        self._tasks.remove(task)

        try:
            result = cast(SoftwareVersionResult, task.finish())
        except Exception as err:
            log.warning("Could not retrieve software version: %s", err)
            result = None

        resource = task.get_user_data()
        device_grid = self._devices[resource]
        device_grid.set_software(result)

    def _set_entity_time(self, task: Task) -> None:
        self._tasks.remove(task)

        try:
            entity_time = cast(str, task.finish())
        except Exception as err:
            log.warning("Could not retrieve entity time: %s", err)
            entity_time = None

        resource = task.get_user_data()

        device_grid = self._devices[resource]
        device_grid.set_entity_time(entity_time)

    def _save_annotation(self) -> None:
        if self._bare_contact is None or not self._is_in_roster:
            return

        if not self._client.is_available():
            return

        buffer_ = self._ui.textview_annotation.get_buffer()
        new_annotation = buffer_.get_property("text")
        note = self._client.get_module("Annotations").get_note(self._bare_contact.jid)
        if note is None or new_annotation != note.data:
            new_note = AnnotationNote(jid=self._bare_contact.jid, data=new_annotation)
            self._client.get_module("Annotations").set_note(new_note)

    def _on_from_subscription_switch_toggled(
        self, switch: Gtk.Switch, state: bool
    ) -> int:
        def _on_response(response_id: str) -> None:
            if response_id == "stop":
                assert self._bare_contact is not None
                self._client.get_module("Presence").unsubscribed(self._bare_contact.jid)
                switch.set_state(state)
            else:
                switch.set_active(True)

        if state:
            assert self._bare_contact is not None
            self._client.get_module("Presence").subscribed(self._bare_contact.jid)
            switch.set_state(state)
        else:
            AlertDialog(
                _("Stop Sharing Online Status?"),
                _(
                    "The contact will be informed that you stopped sharing your "
                    "status. Please note that this can have other side effects."
                ),
                responses=[
                    CancelDialogResponse(),
                    DialogResponse(
                        "stop", _("_Stop Sharing"), appearance="destructive"
                    ),
                ],
                callback=_on_response,
                parent=self,
            )
        return Gdk.EVENT_STOP

    def _on_to_subscription_button_clicked(self, _widget: Gtk.Button) -> None:
        # Save auto_auth if switch for disclosing presence is active
        assert self._bare_contact is not None
        self._client.get_module("Presence").subscribe(
            self._bare_contact.jid,
            auto_auth=self._ui.from_subscription_switch.get_active(),
        )
        self._ui.request_stack.set_visible_child_name("requested")

    def _on_subscribed_presence_received(
        self, _event: SubscribedPresenceReceived
    ) -> None:
        self._ui.to_subscription_stack.set_visible_child_name("checkmark")

    def _on_unsubscribed_presence_received(
        self, _event: UnsubscribedPresenceReceived
    ) -> None:
        self._ui.to_subscription_stack.set_visible_child_name("request")
        self._ui.request_stack.set_visible_child_name("cross")

    def _on_group_selection_changed(self, _widget: Gtk.TreeSelection) -> None:
        selection = self._ui.groups_treeview.get_selection()
        (_model, iter_) = selection.get_selected()
        self._ui.group_remove_button.set_sensitive(bool(iter_))

    def _on_group_name_edited(
        self, _renderer: Gtk.CellRendererText, path: str, new_name: str
    ) -> None:
        if new_name == "":
            return

        model = self._ui.groups_treeview.get_model()
        assert model is not None
        assert isinstance(model, Gtk.ListStore)
        selected_iter_ = model.get_iter(path)
        old_name = model[selected_iter_][Column.GROUP_NAME]

        # Check if group already exists
        iter_ = model.get_iter_first()
        while iter_:
            group_name = model.get_value(iter_, Column.GROUP_NAME)
            if group_name == new_name:
                return
            iter_ = model.iter_next(iter_)

        model.set_value(selected_iter_, Column.GROUP_NAME, new_name)
        self._client.get_module("Roster").rename_group(old_name, new_name)

    def _on_group_remove_button_clicked(self, _widget: Gtk.Button) -> int:
        selection = self._ui.groups_treeview.get_selection()
        model, iter_ = selection.get_selected()
        assert iter_ is not None
        group = model[iter_][Column.GROUP_NAME]

        def _on_response() -> None:
            self._client.get_module("Roster").remove_group(group)
            del model[iter_]

        ConfirmationAlertDialog(
            _("Remove Group?"),
            _('Do you want to remove "%(group)s"?') % {"group": group},
            confirm_label=_("_Remove"),
            appearance="destructive",
            callback=_on_response,
            parent=self,
        )
        return Gdk.EVENT_STOP

    def _on_group_toggled(self, _renderer: Gtk.CellRendererToggle, path: str) -> None:
        assert self._bare_contact is not None
        model = self._ui.groups_treeview.get_model()
        assert model is not None
        model[path][Column.IN_GROUP] = not model[path][Column.IN_GROUP]

        groups: set[str] = set()
        iter_ = model.get_iter_first()
        while iter_:
            if model.get_value(iter_, Column.IN_GROUP):
                group_name = model.get_value(iter_, Column.GROUP_NAME)
                groups.add(group_name)
            iter_ = model.iter_next(iter_)

        self._client.get_module("Roster").set_groups(self._bare_contact.jid, groups)

    def _on_group_add_button_clicked(self, _widget: Gtk.Button) -> None:
        default_name = _("New Group")
        model = self._ui.groups_treeview.get_model()
        assert model is not None
        assert isinstance(model, Gtk.ListStore)

        # Check if default_name group already exists
        iter_ = model.get_iter_first()
        while iter_:
            group_name = model.get_value(iter_, Column.GROUP_NAME)
            if group_name == default_name:
                default_name += "_"
            iter_ = model.iter_next(iter_)

        new_iter_ = model.append([False, default_name])
        path = model.get_path(new_iter_)
        column = self._ui.groups_treeview.get_column(Column.GROUP_NAME)
        self._ui.groups_treeview.set_cursor(path, column, True)


@Gtk.Template(string=get_ui_string("device_info.ui"))
class DeviceInfo(Adw.PreferencesGroup):
    __gtype_name__ = "DeviceInfo"

    _spinner: Adw.Spinner = Gtk.Template.Child()
    _status_row: Adw.ActionRow = Gtk.Template.Child()
    _priority_row: Adw.ActionRow = Gtk.Template.Child()
    _software_row: Adw.ActionRow = Gtk.Template.Child()
    _operating_system_row: Adw.ActionRow = Gtk.Template.Child()
    _local_time_row: Adw.ActionRow = Gtk.Template.Child()

    def __init__(self, contact: ResourceContact) -> None:
        Adw.PreferencesGroup.__init__(self)

        self._contact = contact

        self.set_title(_('Device "%s"') % contact.resource)

        self._status_row.set_subtitle(get_uf_show(contact.show.value))

        self._priority_row.set_subtitle(str(self._contact.priority))
        self._priority_row.set_visible(True)

        self._waiting_for_info = 2

    def run_destroy(self) -> None:
        app.check_finalize(self)

    def set_entity_time(self, entity_time: str | None) -> None:
        if entity_time is not None:
            self._local_time_row.set_subtitle(entity_time)
            self._local_time_row.set_visible(True)

        self._check_complete()

    def set_software(self, software: SoftwareVersionResult | None) -> None:
        if software is not None:
            software_string = f"{software.name} {software.version}"
            self._software_row.set_subtitle(software_string)
            self._software_row.set_visible(True)
            if software.os is not None:
                self._operating_system_row.set_subtitle(software.os)
                self._operating_system_row.set_visible(True)

        self._check_complete()

    def _check_complete(self) -> None:
        self._waiting_for_info -= 1
        if not self._waiting_for_info:
            self._spinner.set_visible(False)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


# Handles Jingle contents (XEP 0166)


from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

from collections.abc import Callable

import nbxmpp
from nbxmpp.namespaces import Namespace

from gajim.common import app
from gajim.common.file_props import FileProp

if TYPE_CHECKING:
    from gajim.common.jingle_session import JingleSession
    from gajim.common.jingle_transport import JingleTransport


contents: dict[str, Any] = {}


def get_jingle_content(node: nbxmpp.Node):
    namespace = node.getNamespace()
    if namespace in contents:
        return contents[namespace](node)


class JingleContentSetupException(Exception):
    """
    Exception that should be raised when a content fails to setup.
    """


class JingleContent:
    """
    An abstraction of content in Jingle sessions
    """

    def __init__(
        self,
        session: JingleSession,
        transport: JingleTransport | None,
        senders: str | None,
    ) -> None:

        self.session = session
        self.transport = transport
        # will be filled by JingleSession.add_content()
        # don't uncomment these lines, we will catch more buggy code then
        # (a JingleContent not added to session shouldn't send anything)
        self.creator: str | None = None
        self.name: str | None = None
        self.accepted = False
        self.sent = False
        self.negotiated = False

        self.media: str | None = None

        self.senders = senders
        if self.senders is None:
            self.senders = "both"

        # Used for stream direction, attribute 'senders'
        self.allow_sending = True

        # These were found by the Politie
        self.file_props: FileProp | None = None

        self.callbacks: dict[
            str,
            list[Callable[[nbxmpp.Node, nbxmpp.Node, nbxmpp.Node | None, str], None]],
        ] = {
            # these are called when *we* get stanzas
            "content-accept": [self.__on_transport_info, self.__on_content_accept],
            "content-add": [self.__on_transport_info],
            "content-modify": [],
            "content-reject": [],
            "content-remove": [],
            "description-info": [],
            "security-info": [],
            "session-accept": [self.__on_transport_info, self.__on_content_accept],
            "session-info": [],
            "session-initiate": [self.__on_transport_info],
            "session-terminate": [],
            "transport-info": [self.__on_transport_info],
            "transport-replace": [self.__on_transport_replace],
            "transport-accept": [self.__on_transport_replace],
            "transport-reject": [],
            "iq-result": [],
            "iq-error": [],
            # these are called when *we* sent these stanzas
            "content-accept-sent": [
                self.__fill_jingle_stanza,
                self.__on_content_accept,
            ],
            "content-add-sent": [self.__fill_jingle_stanza],
            "session-initiate-sent": [self.__fill_jingle_stanza],
            "session-accept-sent": [
                self.__fill_jingle_stanza,
                self.__on_content_accept,
            ],
            "session-terminate-sent": [],
        }

    def is_ready(self) -> bool:
        return self.accepted and not self.sent

    def __on_content_accept(
        self,
        stanza: nbxmpp.Node,
        content: nbxmpp.Node,
        error: nbxmpp.Node | None,
        action: str,
    ) -> None:
        self.on_negotiated()

    def on_negotiated(self) -> None:
        if self.accepted:
            self.negotiated = True
            self.session.content_negotiated(self.media)

    def add_remote_candidates(self, candidates):
        """
        Add a list of candidates to the list of remote candidates
        """
        self.transport.remote_candidates = candidates

    def on_stanza(
        self,
        stanza: nbxmpp.Node,
        content: nbxmpp.Node | None,
        error: nbxmpp.Node | None,
        action: str,
    ) -> None:
        """
        Called when something related to our content was sent by peer
        """
        if action in self.callbacks:
            for callback in self.callbacks[action]:
                callback(stanza, content, error, action)

    def __on_transport_replace(
        self,
        stanza: nbxmpp.Node,
        content: nbxmpp.Node,
        error: nbxmpp.Node | None,
        action: str,
    ) -> None:
        content.addChild(node=self.transport.make_transport())

    def __on_transport_info(
        self,
        stanza: nbxmpp.Node,
        content: nbxmpp.Node,
        error: nbxmpp.Node | None,
        action: str,
    ) -> None:
        """
        Got a new transport candidate
        """
        candidates = self.transport.parse_transport_stanza(content.getTag("transport"))
        if candidates:
            self.add_remote_candidates(candidates)

    def __content(self, payload: list[nbxmpp.Node] | None = None) -> nbxmpp.Node:
        """
        Build a XML content-wrapper for our data
        """
        if payload is None:
            payload = []
        return nbxmpp.Node(
            "content",
            attrs={"name": self.name, "creator": self.creator, "senders": self.senders},
            payload=payload,
        )

    def send_candidate(self, candidate: dict[str, Any]) -> None:
        """
        Send a transport candidate for a previously defined transport.
        """
        content = self.__content()
        content.addChild(node=self.transport.make_transport([candidate]))
        self.session.send_transport_info(content)

    def send_error_candidate(self) -> None:
        """
        Sends a candidate-error when we can't connect to a candidate.
        """
        content = self.__content()
        tp = self.transport.make_transport(add_candidates=False)
        tp.addChild(name="candidate-error")
        content.addChild(node=tp)
        self.session.send_transport_info(content)

    def send_description_info(self) -> None:
        content = self.__content()
        self._fill_content(content)
        self.session.send_description_info(content)

    def __fill_jingle_stanza(
        self,
        stanza: nbxmpp.Node,
        content: nbxmpp.Node,
        error: nbxmpp.Node | None,
        action: str,
    ) -> None:
        """
        Add our things to session-initiate stanza
        """
        self._fill_content(content)
        self.sent = True
        content.addChild(node=self.transport.make_transport())

    def _fill_content(self, content: nbxmpp.Node) -> None:
        description_node = nbxmpp.simplexml.Node(
            tag=Namespace.JINGLE_FILE_TRANSFER_5 + " description"
        )
        file_tag = description_node.setTag("file")
        if self.file_props.name:
            node = nbxmpp.simplexml.Node(tag="name")
            node.addData(self.file_props.name)
            file_tag.addChild(node=node)
        if self.file_props.date:
            node = nbxmpp.simplexml.Node(tag="date")
            node.addData(self.file_props.date)
            file_tag.addChild(node=node)
        if self.file_props.size:
            node = nbxmpp.simplexml.Node(tag="size")
            node.addData(self.file_props.size)
            file_tag.addChild(node=node)
        if self.file_props.type_ == "r":
            if self.file_props.hash_:
                file_tag.addChild(
                    "hash",
                    attrs={"algo": self.file_props.algo},
                    namespace=Namespace.HASHES_2,
                    payload=self.file_props.hash_,
                )
        else:
            # if the file is less than 10 mb, then it is small
            # lets calculate it right away
            if self.file_props.size < 10000000 and not self.file_props.hash_:
                hash_data = self._compute_hash()
                if hash_data:
                    file_tag.addChild(node=hash_data)
                pjid = app.get_jid_without_resource(self.session.peerjid)
                file_info = {
                    "name": self.file_props.name,
                    "file-name": self.file_props.file_name,
                    "hash": self.file_props.hash_,
                    "size": self.file_props.size,
                    "date": self.file_props.date,
                    "peerjid": pjid,
                }
                self.session.connection.get_module("Jingle").set_file_info(file_info)
        desc = file_tag.setTag("desc")
        if self.file_props.desc:
            desc.setData(self.file_props.desc)
        content.addChild(node=description_node)

    def destroy(self) -> None:
        self.callbacks = None
        del self.session.contents[(self.creator, self.name)]



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import NamedTuple

from dataclasses import dataclass
from datetime import datetime
from datetime import UTC
from enum import Enum
from enum import IntEnum
from enum import unique
from functools import total_ordering

from gi.repository import GdkPixbuf
from gi.repository import Gio
from nbxmpp.const import PresenceShow
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from PIL import Image

from gajim.common.i18n import _
from gajim.common.i18n import p_

STOP_EVENT = True
PROPAGATE_EVENT = False

MAX_MESSAGE_CORRECTION_DELAY = 3600


class ValueMissingT:
    pass


VALUE_MISSING = ValueMissingT()


class EncryptionInfoMsg(Enum):
    BAD_MUC_CONFIG = _(
        "This chat’s configuration is unsuitable for "
        "encryption with {encryption}. To use {encryption} in this chat, "
        "it should be non-anonymous and members-only."
    )
    NO_FINGERPRINTS = _(
        "To send an encrypted message, you have to decide "
        "whether to trust the device of your contact."
    )
    QUERY_DEVICES = _(
        "No devices found to encrypt this message to. Querying for devices now…"
    )
    UNDECIDED_FINGERPRINTS = _(
        "There are devices for which you have not made a trust decision yet."
    )
    SESSION_BUILD = _("Successfully built new session with device {device_id}.")


class Entity(NamedTuple):
    jid: JID
    node: str
    hash: str
    method: str


class RowHeaderType(IntEnum):
    CONVERSATIONS = 0
    PINNED = 1


class AvatarSize(IntEnum):
    TAB = 16
    SHOW_CIRCLE = 24
    SMALL = 24
    ROSTER = 32
    START_CHAT = 32
    VCARD_HEADER = 40
    ACCOUNT_SIDE_BAR = 24
    ACCOUNT_SIDE_BAR_WARNING = 20
    WORKSPACE = 36
    WORKSPACE_EDIT = 100
    CHAT = 48
    CHAT_BANNER = 40
    MESSAGE_REPLY = 24
    NOTIFICATION = 96
    CALL = 100
    CALL_BIG = 200
    GROUP_INFO = 100
    TOOLTIP = 72
    ACCOUNT_PAGE = 150
    ACTIVITY_PAGE = 100
    VCARD = 200
    PUBLISH = 200


class ArchiveState:
    NEVER = None
    ALL = datetime(1970, 1, 1, tzinfo=UTC)


@unique
class PathLocation(IntEnum):
    CONFIG = 0
    CACHE = 1
    DATA = 2
    NONE = 3


@unique
class PathType(IntEnum):
    FILE = 0
    FOLDER = 1
    FOLDER_OPTIONAL = 2


@unique
class StyleAttr(Enum):
    COLOR = "color"
    BACKGROUND = "background"
    FONT = "font"


@unique
class CSSPriority(IntEnum):
    APPLICATION = 600
    APPLICATION_DARK = 601
    DEFAULT_THEME = 610
    DEFAULT_THEME_DARK = 611
    USER_THEME = 650


@unique
class ButtonAction(Enum):
    DESTRUCTIVE = "destructive-action"
    SUGGESTED = "suggested-action"


@unique
class IdleState(Enum):
    UNKNOWN = "unknown"
    XA = "xa"
    AWAY = "away"
    AWAKE = "online"


class SyncThreshold(IntEnum):
    NO_SYNC = -1
    NO_THRESHOLD = 0

    def __str__(self):
        return str(self.value)


class MUCUser(IntEnum):
    JID = 0
    NICK = 1
    REASON = 1
    NICK_OR_REASON = 1
    ROLE = 2
    AFFILIATION = 3
    AFFILIATION_TEXT = 4


@unique
class Trust(IntEnum):
    UNTRUSTED = 0
    UNDECIDED = 1
    BLIND = 2
    VERIFIED = 3


class Direction(IntEnum):
    NEXT = 0
    PREV = 1


class Display(Enum):
    X11 = "X11Display"
    WAYLAND = "GdkWaylandDisplay"
    WIN32 = "GdkWin32Display"
    QUARTZ = "GdkQuartzDisplay"


# https://xmpp.org/registrar/querytypes.html
class XmppUriQuery(Enum):
    NONE = ""
    MESSAGE = "message"
    JOIN = "join"

    @staticmethod
    def from_str(s: str) -> XmppUriQuery | None:
        try:
            return XmppUriQuery(s)
        except ValueError:
            return None

    @staticmethod
    def from_str_or_none(s: str) -> XmppUriQuery:
        qtype = XmppUriQuery.from_str(s)
        if qtype:
            return qtype
        return XmppUriQuery.NONE


class MUCJoinedState(Enum):
    JOINED = "joined"
    NOT_JOINED = "not joined"
    BEFORE_JOINING = "before joining"
    JOINING = "joining"
    CREATING = "creating"
    CAPTCHA_REQUEST = "captcha in progress"
    CAPTCHA_FAILED = "captcha failed"
    PASSWORD_REQUEST = "password request"  # noqa: S105

    def __str__(self):
        return self.name

    @property
    def is_joined(self):
        return self == MUCJoinedState.JOINED

    @property
    def is_not_joined(self):
        return self == MUCJoinedState.NOT_JOINED

    @property
    def is_joining(self):
        return self == MUCJoinedState.JOINING

    @property
    def is_before_joining(self):
        return self == MUCJoinedState.BEFORE_JOINING

    @property
    def is_creating(self):
        return self == MUCJoinedState.CREATING

    @property
    def is_captcha_request(self):
        return self == MUCJoinedState.CAPTCHA_REQUEST

    @property
    def is_captcha_failed(self):
        return self == MUCJoinedState.CAPTCHA_FAILED

    @property
    def is_password_request(self):
        return self == MUCJoinedState.PASSWORD_REQUEST


class ClientState(IntEnum):
    DISCONNECTING = 0
    DISCONNECTED = 1
    RECONNECT_SCHEDULED = 2
    CONNECTING = 3
    HOST_META_REQUEST = 4
    CONNECTED = 5
    AVAILABLE = 6

    @property
    def is_disconnecting(self):
        return self == ClientState.DISCONNECTING

    @property
    def is_disconnected(self):
        return self == ClientState.DISCONNECTED

    @property
    def is_reconnect_scheduled(self):
        return self == ClientState.RECONNECT_SCHEDULED

    @property
    def is_connecting(self):
        return self == ClientState.CONNECTING

    @property
    def is_host_meta_request(self):
        return self == ClientState.HOST_META_REQUEST

    @property
    def is_connected(self):
        return self == ClientState.CONNECTED

    @property
    def is_available(self):
        return self == ClientState.AVAILABLE


class SimpleClientState(Enum):
    CONNECTING = "connecting"
    DISCONNECTED = "disconnected"
    CONNECTED = "connected"
    RESUME_IN_PROGRESS = "resume-in-progress"

    @property
    def is_connecting(self):
        return self == SimpleClientState.CONNECTING

    @property
    def is_connected(self):
        return self == SimpleClientState.CONNECTED

    @property
    def is_disconnected(self):
        return self == SimpleClientState.DISCONNECTED

    @property
    def is_resume_in_progress(self):
        return self == SimpleClientState.RESUME_IN_PROGRESS


class JingleState(Enum):
    NULL = "stop"
    CONNECTING = "connecting"
    CONNECTION_RECEIVED = "connection_received"
    CONNECTED = "connected"
    ERROR = "error"

    def __str__(self):
        return self.value


class CallType(Enum):
    AUDIO = "audio"
    VIDEO = "video"


@dataclass
class Draft:
    text: str
    reply_pk: int | None


MUC_DISCO_ERRORS = {
    "remote-server-not-found": _("Remote server not found"),
    "remote-server-timeout": _("Remote server timeout"),
    "service-unavailable": _("Address does not belong to a group chat server"),
    "subscription-required": _("Address does not belong to a group chat server"),
    "not-muc-service": _("Address does not belong to a group chat server"),
    "already-exists": _("Group chat already exists"),
    "item-not-found": _("Group chat does not exist"),
    "gone": _("Group chat is closed"),
}


EME_MESSAGES = {
    "urn:xmpp:otr:0": _(
        "This message was encrypted with OTR and could not be decrypted."
    ),
    "jabber:x:encrypted": _(
        "This message was encrypted with Legacy "
        "OpenPGP and could not be decrypted. You can install "
        "the PGP plugin to handle those messages."
    ),
    "urn:xmpp:openpgp:0": _(
        "This message was encrypted with OpenPGP for XMPP and could not be decrypted."
    ),
    "fallback": _("This message was encrypted with %s and could not be decrypted."),
}


EME_PROTOCOLS = {
    "urn:xmpp:otr:0": "OTR",
    "jabber:x:encrypted": "PGP",
    "urn:xmpp:openpgp:0": "OpenPGP",
    "eu.siacs.conversations.axolotl": "OMEMO",
    "urn:xmpp:omemo:1": "OMEMO",
    "urn:xmpp:omemo:2": "OMEMO",
}


LOCATION_DATA = {
    "accuracy": _("accuracy"),
    "alt": _("alt"),
    "area": _("area"),
    "bearing": _("bearing"),
    "building": _("building"),
    "country": _("country"),
    "countrycode": _("countrycode"),
    "datum": _("datum"),
    "description": _("description"),
    "error": _("error"),
    "floor": _("floor"),
    "lat": _("lat"),
    "locality": _("locality"),
    "lon": _("lon"),
    "postalcode": _("postalcode"),
    "region": _("region"),
    "room": _("room"),
    "speed": _("speed"),
    "street": _("street"),
    "text": _("text"),
    "timestamp": _("timestamp"),
    "uri": _("URI"),
}


SSLError = {
    2: _("Unable to get issuer certificate"),
    3: _("Unable to get certificate CRL"),
    4: _("Unable to decrypt certificate's signature"),
    5: _("Unable to decrypt CRL's signature"),
    6: _("Unable to decode issuer public key"),
    7: _("Certificate signature failure"),
    8: _("CRL signature failure"),
    9: _("Certificate is not yet valid"),
    10: _("Certificate has expired"),
    11: _("CRL is not yet valid"),
    12: _("CRL has expired"),
    13: _("Format error in certificate's notBefore field"),
    14: _("Format error in certificate's notAfter field"),
    15: _("Format error in CRL's lastUpdate field"),
    16: _("Format error in CRL's nextUpdate field"),
    17: _("Out of memory"),
    18: _("Self signed certificate"),
    19: _("Self signed certificate in certificate chain"),
    20: _("Unable to get local issuer certificate"),
    21: _("Unable to verify the first certificate"),
    22: _("Certificate chain too long"),
    23: _("Certificate revoked"),
    24: _("Invalid CA certificate"),
    25: _("Path length constraint exceeded"),
    26: _("Unsupported certificate purpose"),
    27: _("Certificate not trusted"),
    28: _("Certificate rejected"),
    29: _("Subject issuer mismatch"),
    30: _("Authority and subject key identifier mismatch"),
    31: _("Authority and issuer serial number mismatch"),
    32: _("Key usage does not include certificate signing"),
    50: _("Application verification failure"),
}


VOWELS = "aeiou"


CONSONANTS = "bcdfghjklmnpqrstvwxyz"


THANKS = [
    "Alexander Futász",
    "Alexander V. Butenko",
    "Alexey Nezhdanov",
    "Alfredo Junix",
    "Anaël Verrier",
    "Anders Ström",
    "Andrew Sayman",
    "Anton Shmigirilov",
    "Christian Bjälevik",
    "Christophe Got",
    "Christoph Neuroth",
    "David Campey",
    "Dennis Craven",
    "Fabian Neumann",
    "Filippos Papadopoulos",
    "Francisco Alburquerque Parra (Membris Khan)",
    "Frederic Lory",
    "Fridtjof Bussefor",
    "Geobert Quach",
    "Guillaume Morin",
    "Gustavo J. A. M. Carneiro",
    "Ivo Anjo",
    "Josef Vybíral",
    "Juraj Michalek",
    "Kjell Braden",
    "Luis Peralta",
    "Michael Scherer",
    "Michele Campeotto",
    "Mike Albon",
    "Miguel Fonseca",
    "Norman Rasmussen",
    "Oscar Hellström",
    "Peter Saint-Andre",
    "Petr Menšík",
    "Sergey Kuleshov",
    "Stavros Giannouris",
    "Stian B. Barmen",
    "Thilo Molitor",
    "Thomas Klein-Hitpaß",
    "Urtzi Alfaro",
    "Witold Kieraś",
    "Yakov Bezrukov",
    "Yavor Doganov",
]


ARTISTS = [
    "Anders Ström",
    "Christophe Got",
    "Dennis Craven",
    "Dmitry Korzhevin",
    "Guillaume Morin",
    "Gvorcek Spajreh",
    "Josef Vybíral",
    "Membris Khan",
    "Rederick Asher",
    "Jakub Szypulka",
]


MAINTAINERS = [
    "Philipp Hörist (philipp AT hoerist.com)",
    "Daniel Brötzmann (cal0pteryx AT posteo.de)",
]


DEVELOPERS = [
    "Marc Henning",
    "André Apitzsch",
    "Yann Leboulanger (asterix AT lagaule.org)",
    "Stefan Bethge (stefan AT lanpartei.de)",
    "Alexander Cherniuk (ts33kr AT gmail.com)",
    "Stephan Erb (steve-e AT h3c.de)",
    "Vincent Hanquez (tab AT snarc.org)",
    "Dimitur Kirov (dkirov AT gmail.com)",
    "Nikos Kouremenos (kourem AT gmail.com)",
    "Julien Pivotto (roidelapluie AT gmail.com)",
    "Jonathan Schleifer (js-gajim AT webkeks.org)",
    "Travis Shirk (travis AT pobox.com)",
    "Brendan Taylor (whateley AT gmail.com)",
    "Jean-Marie Traissard (jim AT lapin.org)",
]


RFC5646_LANGUAGE_TAGS = {
    "af": "Afrikaans",
    "af-ZA": "Afrikaans (South Africa)",
    "ar": "Arabic",
    "ar-AE": "Arabic (U.A.E.)",
    "ar-BH": "Arabic (Bahrain)",
    "ar-DZ": "Arabic (Algeria)",
    "ar-EG": "Arabic (Egypt)",
    "ar-IQ": "Arabic (Iraq)",
    "ar-JO": "Arabic (Jordan)",
    "ar-KW": "Arabic (Kuwait)",
    "ar-LB": "Arabic (Lebanon)",
    "ar-LY": "Arabic (Libya)",
    "ar-MA": "Arabic (Morocco)",
    "ar-OM": "Arabic (Oman)",
    "ar-QA": "Arabic (Qatar)",
    "ar-SA": "Arabic (Saudi Arabia)",
    "ar-SY": "Arabic (Syria)",
    "ar-TN": "Arabic (Tunisia)",
    "ar-YE": "Arabic (Yemen)",
    "az": "Azeri (Latin)",
    "az-AZ": "Azeri (Latin) (Azerbaijan)",
    "az-Cyrl-AZ": "Azeri (Cyrillic) (Azerbaijan)",
    "be": "Belarusian",
    "be-BY": "Belarusian (Belarus)",
    "bg": "Bulgarian",
    "bg-BG": "Bulgarian (Bulgaria)",
    "bs-BA": "Bosnian (Bosnia and Herzegovina)",
    "ca": "Catalan",
    "ca-ES": "Catalan (Spain)",
    "cs": "Czech",
    "cs-CZ": "Czech (Czech Republic)",
    "cy": "Welsh",
    "cy-GB": "Welsh (United Kingdom)",
    "da": "Danish",
    "da-DK": "Danish (Denmark)",
    "de": "German",
    "de-AT": "German (Austria)",
    "de-CH": "German (Switzerland)",
    "de-DE": "German (Germany)",
    "de-LI": "German (Liechtenstein)",
    "de-LU": "German (Luxembourg)",
    "dv": "Divehi",
    "dv-MV": "Divehi (Maldives)",
    "el": "Greek",
    "el-GR": "Greek (Greece)",
    "en": "English",
    "en-AU": "English (Australia)",
    "en-BZ": "English (Belize)",
    "en-CA": "English (Canada)",
    "en-CB": "English (Caribbean)",
    "en-GB": "English (United Kingdom)",
    "en-IE": "English (Ireland)",
    "en-JM": "English (Jamaica)",
    "en-NZ": "English (New Zealand)",
    "en-PH": "English (Republic of the Philippines)",
    "en-TT": "English (Trinidad and Tobago)",
    "en-US": "English (United States)",
    "en-ZA": "English (South Africa)",
    "en-ZW": "English (Zimbabwe)",
    "eo": "Esperanto",
    "es": "Spanish",
    "es-AR": "Spanish (Argentina)",
    "es-BO": "Spanish (Bolivia)",
    "es-CL": "Spanish (Chile)",
    "es-CO": "Spanish (Colombia)",
    "es-CR": "Spanish (Costa Rica)",
    "es-DO": "Spanish (Dominican Republic)",
    "es-EC": "Spanish (Ecuador)",
    "es-ES": "Spanish (Spain)",
    "es-GT": "Spanish (Guatemala)",
    "es-HN": "Spanish (Honduras)",
    "es-MX": "Spanish (Mexico)",
    "es-NI": "Spanish (Nicaragua)",
    "es-PA": "Spanish (Panama)",
    "es-PE": "Spanish (Peru)",
    "es-PR": "Spanish (Puerto Rico)",
    "es-PY": "Spanish (Paraguay)",
    "es-SV": "Spanish (El Salvador)",
    "es-UY": "Spanish (Uruguay)",
    "es-VE": "Spanish (Venezuela)",
    "et": "Estonian",
    "et-EE": "Estonian (Estonia)",
    "eu": "Basque",
    "eu-ES": "Basque (Spain)",
    "fa": "Farsi",
    "fa-IR": "Farsi (Iran)",
    "fi": "Finnish",
    "fi-FI": "Finnish (Finland)",
    "fo": "Faroese",  # codespell:ignore
    "fo-FO": "Faroese (Faroe Islands)",
    "fr": "French",
    "fr-BE": "French (Belgium)",
    "fr-CA": "French (Canada)",
    "fr-CH": "French (Switzerland)",
    "fr-FR": "French (France)",
    "fr-LU": "French (Luxembourg)",
    "fr-MC": "French (Principality of Monaco)",
    "gl": "Galician",
    "gl-ES": "Galician (Spain)",
    "gu": "Gujarati",
    "gu-IN": "Gujarati (India)",
    "he": "Hebrew",
    "he-IL": "Hebrew (Israel)",
    "hi": "Hindi",
    "hi-IN": "Hindi (India)",
    "hr": "Croatian",
    "hr-BA": "Croatian (Bosnia and Herzegovina)",
    "hr-HR": "Croatian (Croatia)",
    "hu": "Hungarian",
    "hu-HU": "Hungarian (Hungary)",
    "hy": "Armenian",
    "hy-AM": "Armenian (Armenia)",
    "id": "Indonesian",
    "id-ID": "Indonesian (Indonesia)",
    "is": "Icelandic",
    "is-IS": "Icelandic (Iceland)",
    "it": "Italian",
    "it-CH": "Italian (Switzerland)",
    "it-IT": "Italian (Italy)",
    "ja": "Japanese",
    "ja-JP": "Japanese (Japan)",
    "ka": "Georgian",
    "ka-GE": "Georgian (Georgia)",
    "kk": "Kazakh",
    "kk-KZ": "Kazakh (Kazakhstan)",
    "kn": "Kannada",
    "kn-IN": "Kannada (India)",
    "ko": "Korean",
    "ko-KR": "Korean (Korea)",
    "kok": "Konkani",
    "kok-IN": "Konkani (India)",
    "ky": "Kyrgyz",
    "ky-KG": "Kyrgyz (Kyrgyzstan)",
    "lt": "Lithuanian",
    "lt-LT": "Lithuanian (Lithuania)",
    "lv": "Latvian",
    "lv-LV": "Latvian (Latvia)",
    "mi": "Maori",
    "mi-NZ": "Maori (New Zealand)",
    "mk": "FYRO Macedonian",
    "mk-MK": "FYRO Macedonian (Former Yugoslav Republic of Macedonia)",
    "mn": "Mongolian",
    "mn-MN": "Mongolian (Mongolia)",
    "mr": "Marathi",
    "mr-IN": "Marathi (India)",
    "ms": "Malay",
    "ms-BN": "Malay (Brunei Darussalam)",
    "ms-MY": "Malay (Malaysia)",
    "mt": "Maltese",
    "mt-MT": "Maltese (Malta)",
    "nb": "Norwegian (Bokm?l)",
    "nb-NO": "Norwegian (Bokm?l) (Norway)",
    "nl": "Dutch",
    "nl-BE": "Dutch (Belgium)",
    "nl-NL": "Dutch (Netherlands)",
    "nn-NO": "Norwegian (Nynorsk) (Norway)",
    "ns": "Northern Sotho",
    "ns-ZA": "Northern Sotho (South Africa)",
    "pa": "Punjabi",
    "pa-IN": "Punjabi (India)",
    "pl": "Polish",
    "pl-PL": "Polish (Poland)",
    "ps": "Pashto",
    "ps-AR": "Pashto (Afghanistan)",
    "pt": "Portuguese",
    "pt-BR": "Portuguese (Brazil)",
    "pt-PT": "Portuguese (Portugal)",
    "qu": "Quechua",
    "qu-BO": "Quechua (Bolivia)",
    "qu-EC": "Quechua (Ecuador)",
    "qu-PE": "Quechua (Peru)",
    "ro": "Romanian",
    "ro-RO": "Romanian (Romania)",
    "ru": "Russian",
    "ru-RU": "Russian (Russia)",
    "sa": "Sanskrit",
    "sa-IN": "Sanskrit (India)",
    "se": "Sami",
    "se-FI": "Sami (Finland)",
    "se-NO": "Sami (Norway)",
    "se-SE": "Sami (Sweden)",
    "sk": "Slovak",
    "sk-SK": "Slovak (Slovakia)",
    "sl": "Slovenian",
    "sl-SI": "Slovenian (Slovenia)",
    "sq": "Albanian",
    "sq-AL": "Albanian (Albania)",
    "sr-BA": "Serbian (Latin) (Bosnia and Herzegovina)",
    "sr-Cyrl-BA": "Serbian (Cyrillic) (Bosnia and Herzegovina)",
    "sr-SP": "Serbian (Latin) (Serbia and Montenegro)",
    "sr-Cyrl-SP": "Serbian (Cyrillic) (Serbia and Montenegro)",
    "sv": "Swedish",
    "sv-FI": "Swedish (Finland)",
    "sv-SE": "Swedish (Sweden)",
    "sw": "Swahili",
    "sw-KE": "Swahili (Kenya)",
    "syr": "Syriac",
    "syr-SY": "Syriac (Syria)",
    "ta": "Tamil",
    "ta-IN": "Tamil (India)",
    "te": "Telugu",  # codespell:ignore
    "te-IN": "Telugu (India)",
    "th": "Thai",
    "th-TH": "Thai (Thailand)",
    "tl": "Tagalog",
    "tl-PH": "Tagalog (Philippines)",
    "tn": "Tswana",
    "tn-ZA": "Tswana (South Africa)",
    "tr": "Turkish",
    "tr-TR": "Turkish (Turkey)",
    "tt": "Tatar",
    "tt-RU": "Tatar (Russia)",
    "ts": "Tsonga",
    "uk": "Ukrainian",
    "uk-UA": "Ukrainian (Ukraine)",
    "ur": "Urdu",
    "ur-PK": "Urdu (Islamic Republic of Pakistan)",
    "uz": "Uzbek (Latin)",
    "uz-UZ": "Uzbek (Latin) (Uzbekistan)",
    "uz-Cyrl-UZ": "Uzbek (Cyrillic) (Uzbekistan)",
    "vi": "Vietnamese",
    "vi-VN": "Vietnamese (Vietnam)",
    "xh": "Xhosa",
    "xh-ZA": "Xhosa (South Africa)",
    "zh": "Chinese",
    "zh-CN": "Chinese (S)",
    "zh-HK": "Chinese (Hong Kong)",
    "zh-MO": "Chinese (Macau)",
    "zh-SG": "Chinese (Singapore)",
    "zh-TW": "Chinese (T)",
    "zu": "Zulu",
    "zu-ZA": "Zulu (South Africa)",
}


GIO_TLS_ERRORS = {
    Gio.TlsCertificateFlags.UNKNOWN_CA: _(
        "The signing certificate authority is not known"
    ),
    Gio.TlsCertificateFlags.REVOKED: _("The certificate has been revoked"),
    Gio.TlsCertificateFlags.BAD_IDENTITY: _(
        "The certificate does not match the expected identity of the site"
    ),
    Gio.TlsCertificateFlags.INSECURE: _("The certificate’s algorithm is insecure"),
    Gio.TlsCertificateFlags.NOT_ACTIVATED: _(
        "The certificate’s activation time is in the future"
    ),
    Gio.TlsCertificateFlags.GENERIC_ERROR: _("Unknown validation error"),
    Gio.TlsCertificateFlags.EXPIRED: _("The certificate has expired"),
}


class FTState(Enum):
    INIT = "init"
    PREPARING = "prepare"
    ENCRYPTING = "encrypting"
    DECRYPTING = "decrypting"
    STARTED = "started"
    IN_PROGRESS = "progress"
    FINISHED = "finished"
    ERROR = "error"
    CANCELLED = "cancelled"

    @property
    def is_preparing(self):
        return self == FTState.PREPARING

    @property
    def is_encrypting(self):
        return self == FTState.ENCRYPTING

    @property
    def is_decrypting(self):
        return self == FTState.DECRYPTING

    @property
    def is_started(self):
        return self == FTState.STARTED

    @property
    def is_in_progress(self):
        return self == FTState.IN_PROGRESS

    @property
    def is_finished(self):
        return self == FTState.FINISHED

    @property
    def is_error(self):
        return self == FTState.ERROR

    @property
    def is_cancelled(self):
        return self == FTState.CANCELLED

    @property
    def is_active(self):
        return not (self.is_error or self.is_cancelled or self.is_finished)


SASL_ERRORS = {
    "aborted": _("Authentication aborted"),
    "account-disabled": _("Account disabled"),
    "credentials-expired": _("Credentials expired"),
    "encryption-required": _("Encryption required"),
    "incorrect-encoding": _("Authentication failed"),
    "invalid-authzid": _("Authentication failed"),
    "malformed-request": _("Authentication failed"),
    "invalid-mechanism": _("Authentication mechanism not supported"),
    "mechanism-too-weak": _("Authentication mechanism too weak"),
    "not-authorized": _("Authentication failed"),
    "temporary-auth-failure": _("Authentication currently not possible"),
}


COMMON_FEATURES = [
    Namespace.BYTESTREAM,
    Namespace.MUC,
    Namespace.COMMANDS,
    Namespace.DISCO_INFO,
    Namespace.LAST,
    Namespace.DATA,
    Namespace.ENCRYPTED,
    Namespace.PING,
    Namespace.CHATSTATES,
    Namespace.RECEIPTS,
    Namespace.TIME_REVISED,
    Namespace.VERSION,
    Namespace.ROSTERX,
    Namespace.SECLABEL,
    Namespace.CORRECT,
    Namespace.CHATMARKERS,
    Namespace.EME,
    Namespace.HASHES_2,
    Namespace.HASHES_MD5,
    Namespace.HASHES_SHA1,
    Namespace.HASHES_SHA256,
    Namespace.HASHES_SHA512,
    Namespace.HASHES_SHA3_256,
    Namespace.HASHES_SHA3_512,
    Namespace.HASHES_BLAKE2B_256,
    Namespace.HASHES_BLAKE2B_512,
    # TODO Jingle FT
    # Namespace.JINGLE,
    # Namespace.JINGLE_FILE_TRANSFER_5,
    # Namespace.JINGLE_BYTESTREAM,
    # Namespace.JINGLE_IBB,
    Namespace.AVATAR_METADATA + "+notify",
    Namespace.MESSAGE_MODERATE,
    Namespace.REPLY,
    Namespace.OMEMO_TEMP_DL + "+notify",
    Namespace.OPENPGP_PK + "+notify",
    Namespace.STYLING,
    Namespace.REACTIONS,
    Namespace.MDS + "+notify",
]


SHOW_LIST = [
    "offline",
    "connecting",
    "away",
    "xa",
    "chat",
    "online",
    "dnd",
]

SHOW_STRING = {
    "dnd": p_("User status", "Busy"),
    "xa": p_("User status", "Not Available"),
    "chat": p_("User status", "Free for Chat"),
    "online": p_("User status", "Available"),
    "connecting": _("Connecting"),
    "away": p_("User status", "Away"),
    "offline": _("Offline"),
}

SHOW_STRING_MNEMONIC = {
    "dnd": p_("User status", "_Busy"),
    "xa": p_("User status", "_Not Available"),
    "chat": p_("User status", "_Free for Chat"),
    "online": p_("User status", "_Available"),
    "connecting": _("Connecting"),
    "away": p_("User status", "A_way"),
    "offline": _("_Offline"),
}

GAJIM_FAQ_URI = "https://dev.gajim.org/gajim/gajim/wikis/help/gajimfaq"
GAJIM_PRIVACY_POLICY_URI = "https://gajim.org/privacy/"
GAJIM_WIKI_URI = "https://dev.gajim.org/gajim/gajim/wikis"
GAJIM_SUPPORT_JID = "gajim@conference.gajim.org"


# Clickable URI schemes not (yet) registered with IANA (see ./iana/uri_schemes.py)
NONREGISTERED_URI_SCHEMES = {
    "gemini",
}

# This is an excerpt of Media Types from
# https://www.iana.org/assignments/media-types/media-types.xhtml
# plus some additions
OTHER_MIME_TYPES = {
    # application/
    "application/calendar+json",
    "application/calendar+xml",
    "application/epub+zip",
    "application/json",
    "application/mp4",
    "application/msword",
    "application/ogg",
    "application/pdf",
    "application/pgp-encrypted",
    "application/pgp-signature",
    "application/postscript",
    "application/rtf",
    "application/vcard+json",
    "application/vcard+xml",
    "application/vnd.amazon.mobi8-ebook",
    "application/vnd.google-earth.kml+xml",
    "application/vnd.google-earth.kmz",
    # Start office
    "application/vnd.ms-access",
    "application/vnd.ms-excel",
    "application/vnd.ms-excel.addin.macroEnabled.12",
    "application/vnd.ms-excel.sheet.binary.macroEnabled.12",
    "application/vnd.ms-excel.sheet.macroEnabled.12",
    "application/vnd.ms-excel.template.macroEnabled.12",
    "application/vnd.ms-powerpoint",
    "application/vnd.ms-powerpoint.addin.macroEnabled.12",
    "application/vnd.ms-powerpoint.presentation.macroEnabled.12",
    "application/vnd.ms-powerpoint.slideshow.macroEnabled.12",
    "application/vnd.ms-powerpoint.template.macroEnabled.12",
    "application/vnd.ms-word.document.macroEnabled.12",
    "application/vnd.ms-word.template.macroEnabled.12",
    "application/vnd.openxmlformats-officedocument.presentationml.presentation",
    "application/vnd.openxmlformats-officedocument.presentationml.slideshow",
    "application/vnd.openxmlformats-officedocument.presentationml.template",
    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    "application/vnd.openxmlformats-officedocument.spreadsheetml.template",
    "application/vnd.openxmlformats-officedocument.vmlDrawing",
    "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
    "application/vnd.openxmlformats-officedocument.wordprocessingml.template",
    # End office
    "application/vnd.sqlite3",
    "application/zip",
    # font/*
    "font/ttf",
    "font/woff",
    "font/woff2",
    # model/*
    "model/mtl",
    "model/obj",
    "model/stl",
    # text/*
    "text/calendar",
    "text/csv",
    "text/markdown",
    "text/rtf",
    "text/vcard",
}

AUDIO_MIME_TYPES = {
    "audio/aac",
    "audio/ac3",
    "audio/flac",
    "audio/mp3",
    "audio/mp4",
    "audio/mpeg",
    "audio/ogg",
    "audio/opus",
    "audio/vnd.wave",
    "audio/wav",
    "audio/wave",
    "audio/x-aac",
    "audio/x-flac",
    "audio/x-m4a",
    "audio/x-matroska",
    "audio/x-mp3",
    "audio/x-mpeg",
    "audio/x-ogg",
    "audio/x-wav",
}

VIDEO_MIME_TYPES = {
    "video/H264",
    "video/H265",
    "video/mp4",
    "video/mpeg4-generic",
    "video/ogg",
    "video/quicktime",
    "video/vc1",
    "video/VP8",
    "video/webm",
    "video/x-matroska",
    "video/x-msvideo",
}

# Executable files extensions:
# List from https://source.chromium.org/chromium/chromium/src/+/main:components/safe_browsing/content/resources/download_file_types.asciipb
WINDOWS_EXECUTABLE_EXTENSIONS = {
    "app",
    "bat",
    "bgi",
    "com",
    "cpi",
    "cpl",
    "dll",
    "drv",
    "exe",
    "fon",
    "hta",
    "iqy",
    "js",
    "msc",
    "pif",
    "sys",
}
MACOS_EXECUTABLE_EXTENSIONS = {
    "cdr",
    "dart",
    "dc42",
    "diskcopy42",
    "dmg",
    "dmgpart",
    "dvdr",
    "dylib",
    "img",
    "imgparg",
    "ndif",
    "service",
    "smi",
    "sparsebundle",
    "sparseimage",
    "toast",
    "udif",
}
LINUX_EXECUTABLE_EXTENSIONS = {
    "deb",
    "pet",
    "pkg",
    "pup",
    "rpm",
    "slp",
}


def get_pixbuf_mime_types() -> set[str]:
    pixbuf_mime_types: set[str] = set()

    for fmt in GdkPixbuf.Pixbuf.get_formats():
        mime_types = fmt.get_mime_types()
        if mime_types is None:
            continue
        for mime_type in mime_types:
            pixbuf_mime_types.add(mime_type.lower())

    return pixbuf_mime_types


def get_pil_mime_types() -> set[str]:
    pil_mime_types: set[str] = set()

    Image.init()
    for mime_type in Image.MIME.values():
        pil_mime_types.add(mime_type.lower())

    return set(filter(lambda mime_type: mime_type.startswith("image"), pil_mime_types))


PIL_IMAGE_MIME_TYPES = get_pil_mime_types()
PIXBUF_IMAGE_MIME_TYPES = get_pixbuf_mime_types()
IMAGE_MIME_TYPES = PIL_IMAGE_MIME_TYPES | PIXBUF_IMAGE_MIME_TYPES

ALL_MIME_TYPES = (
    OTHER_MIME_TYPES | IMAGE_MIME_TYPES | AUDIO_MIME_TYPES | VIDEO_MIME_TYPES
)

TRUST_SYMBOL_DATA = {
    Trust.UNTRUSTED: ("lucide-circle-x-symbolic", _("Untrusted"), "error-color"),
    Trust.UNDECIDED: (
        "lucide-shield-alert-symbolic",
        _("Trust Not Decided"),
        "warning-color",
    ),
    Trust.BLIND: (
        "lucide-shield-question-mark-symbolic",
        _("Blind Trust"),
        "encrypted-color",
    ),
    Trust.VERIFIED: ("lucide-shield-check-symbolic", _("Verified"), "encrypted-color"),
}


THRESHOLD_OPTIONS = {
    -1: _("No Sync"),
    1: _("1 Day"),
    2: _("2 Days"),
    7: _("1 Week"),
    30: _("1 Month"),
    0: _("No Threshold"),
}


@total_ordering
class PresenceShowExt(Enum):
    """
    This extends nbxmpp.const.PresenceShow for convenience
    with an OFFLINE member
    """

    OFFLINE = "offline"

    @property
    def is_offline(self):
        return self == PresenceShowExt.OFFLINE

    def __lt__(self, other: PresenceShowExt | PresenceShow) -> bool:
        if isinstance(other, PresenceShowExt):
            return False
        if not isinstance(other, PresenceShow):  # type: ignore
            return NotImplemented
        return True


# Replace keys with Gio.TlsProtocolVersion enum once
# we require GLib >= 2.69
TLS_VERSION_STRINGS = {
    3: "1.1",
    4: "1.2",
    5: "1.3",
}


RETRACTION_FALLBACK = (
    "/me retracted a previous message, but it's unsupported by your client."
)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import logging

import nbxmpp
from nbxmpp.namespaces import Namespace

from gajim.common import app
from gajim.common import types
from gajim.common.jingle_transport import TransportType

# from gajim.common.socks5 import Socks5ReceiverClient
# from gajim.common.socks5 import Socks5SenderClient

if TYPE_CHECKING:
    from gajim.common.jingle_ft import JingleFileTransfer

log = logging.getLogger("gajim.c.jingle_ftstates")


class JingleFileTransferStates:
    """
    This class implements the state machine design pattern
    """

    def __init__(self, jingleft: JingleFileTransfer) -> None:
        self.jft = jingleft

    def action(self, args: dict[str, Any] | None = None) -> None:
        """
        This method MUST be overridden by a subclass
        """
        raise NotImplementedError("This is an abstract method!")


class StateInitialized(JingleFileTransferStates):
    """
    This state initializes the file transfer
    """

    def action(self, args: dict[str, Any] | None = None) -> None:
        if self.jft.weinitiate:
            # update connection's fileprops
            self.jft._listen_host()
            # Listen on configured port for file transfer
        else:
            # Connect to the candidate host, on success call on_connect method
            app.socks5queue.connect_to_hosts(
                self.jft.session.connection.name,
                self.jft.file_props.transport_sid,
                self.jft.on_connect,
                self.jft._on_connect_error,
            )


class StateCandSent(JingleFileTransferStates):
    """
    This state sends our nominated candidate
    """

    def _send_candidate(self, args: dict[str, Any]) -> None:
        if "candError" in args:
            self.jft.nominated_cand["our-cand"] = False
            self.jft.send_error_candidate()
            return
        # Send candidate used
        streamhost = args["streamhost"]
        self.jft.nominated_cand["our-cand"] = streamhost
        content = nbxmpp.Node("content")
        content.setAttr("creator", "initiator")
        content.setAttr("name", self.jft.name)
        transport = nbxmpp.Node("transport")
        transport.setNamespace(Namespace.JINGLE_BYTESTREAM)
        transport.setAttr("sid", self.jft.transport.sid)
        candidateused = nbxmpp.Node("candidate-used")
        candidateused.setAttr("cid", streamhost["candidate_id"])
        transport.addChild(node=candidateused)
        content.addChild(node=transport)
        self.jft.session.send_transport_info(content)

    def action(self, args: dict[str, Any] | None = None) -> None:
        self._send_candidate(args)


class StateCandReceived(JingleFileTransferStates):
    """
    This state happens when we receive a candidate.
    It takes the arguments: canError if we receive a candidate-error
    """

    def _recv_candidate(self, args: dict[str, Any]) -> None:
        if "candError" in args:
            return
        content = args["content"]
        streamhost_cid = (
            content.getTag("transport").getTag("candidate-used").getAttr("cid")
        )
        streamhost_used = None
        for cand in self.jft.transport.candidates:
            if cand["candidate_id"] == streamhost_cid:
                streamhost_used = cand
                break
        if streamhost_used is None:
            log.info("unknown streamhost")
            return
        # We save the candidate nominated by peer
        self.jft.nominated_cand["peer-cand"] = streamhost_used

    def action(self, args: dict[str, Any] | None = None) -> None:
        self._recv_candidate(args)


class StateCandSentAndRecv(StateCandSent, StateCandReceived):
    """
    This state happens when we have received and sent the candidates.
    It takes the boolean argument: sendCand in order to decide whether
    we should execute the action of when we receive or send a candidate.
    """

    def action(self, args: dict[str, Any] | None = None) -> None:
        if args["sendCand"]:
            self._send_candidate(args)
        else:
            self._recv_candidate(args)


class StateTransportReplace(JingleFileTransferStates):
    """
    This state initiates transport replace
    """

    def action(self, args: dict[str, Any] | None = None) -> None:
        self.jft.session.transport_replace()


class StateTransfering(JingleFileTransferStates):
    """
    This state will start the transfer depending on the type of transport
    we have.
    """

    def _start_ibb_transfer(self, con: types.Client) -> None:
        self.jft.file_props.transport_sid = self.jft.transport.sid
        fp = open(self.jft.file_props.file_name, "rb")
        con.get_module("IBB").send_open(
            self.jft.session.peerjid, self.jft.file_props.sid, fp
        )

    def _start_sock5_transfer(self) -> None:
        # It tells whether we start the transfer as client or server
        mode = None
        if self.jft.is_our_candidate_used():
            mode = "client"
            streamhost_used = self.jft.nominated_cand["our-cand"]
            app.socks5queue.remove_server(self.jft.file_props.transport_sid)
        else:
            mode = "server"
            streamhost_used = self.jft.nominated_cand["peer-cand"]
            app.socks5queue.remove_client(self.jft.file_props.transport_sid)
            app.socks5queue.remove_other_servers(streamhost_used["host"])
        if streamhost_used["type"] == "proxy":
            self.jft.file_props.is_a_proxy = True
            if self.jft.file_props.type_ == "s" and self.jft.weinitiate:
                self.jft.file_props.proxy_sender = streamhost_used["initiator"]
                self.jft.file_props.proxy_receiver = streamhost_used["target"]
            else:
                self.jft.file_props.proxy_sender = streamhost_used["target"]
                self.jft.file_props.proxy_receiver = streamhost_used["initiator"]
            if self.jft.file_props.type_ == "s":
                s = app.socks5queue.senders
                for v in s.values():
                    if v.host == streamhost_used["host"] and v.connected:
                        return
            elif self.jft.file_props.type_ == "r":
                r = app.socks5queue.readers
                for v in r.values():
                    if v.host == streamhost_used["host"] and v.connected:
                        return
            else:
                raise TypeError
            self.jft.file_props.streamhost_used = True
            streamhost_used["sid"] = self.jft.file_props.transport_sid
            self.jft.file_props.streamhosts = []
            self.jft.file_props.streamhosts.append(streamhost_used)
            self.jft.file_props.proxyhosts = []
            self.jft.file_props.proxyhosts.append(streamhost_used)
            if self.jft.file_props.type_ == "s":
                app.socks5queue.idx += 1
                idx = app.socks5queue.idx
                sockobj = Socks5SenderClient(
                    app.idlequeue,
                    idx,
                    app.socks5queue,
                    _sock=None,
                    host=str(streamhost_used["host"]),
                    port=int(streamhost_used["port"]),
                    connected=False,
                    file_props=self.jft.file_props,
                )
            else:
                sockobj = Socks5ReceiverClient(
                    app.idlequeue,
                    streamhost_used,
                    transport_sid=self.jft.file_props.transport_sid,
                    file_props=self.jft.file_props,
                )
            sockobj.proxy = True
            sockobj.streamhost = streamhost_used
            app.socks5queue.add_sockobj(self.jft.session.connection.name, sockobj)
            streamhost_used["idx"] = sockobj.queue_idx
            # If we offered the nominated candidate used, we activate
            # the proxy
            if not self.jft.is_our_candidate_used():
                app.socks5queue.on_success[self.jft.file_props.transport_sid] = (
                    self.jft.transport._on_proxy_auth_ok
                )
            # TODO: add on failure
        else:
            app.socks5queue.send_file(
                self.jft.file_props, self.jft.session.connection.name, mode
            )

    def action(self, args: dict[str, Any] | None = None) -> None:
        if self.jft.transport.type_ == TransportType.IBB:
            self._start_ibb_transfer(self.jft.session.connection)
        elif self.jft.transport.type_ == TransportType.SOCKS5:
            self._start_sock5_transfer()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


from gi.repository import Gtk
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import i18n
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant

from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.builder import get_builder
from gajim.gtk.window import GajimAppWindow


class RosterItemExchange(GajimAppWindow):
    """
    Used when someone sends a Roster Item Exchange suggestion (XEP-0144)
    """

    def __init__(
        self,
        account: str,
        action: str,
        exchange_list: dict[str, list[str]],
        jid_from: JID,
        message_body: str | None = None,
    ) -> None:
        GajimAppWindow.__init__(
            self,
            name="RosterItemExchange",
            title=_("Contact List Exchange"),
            add_window_padding=True,
            header_bar=True,
        )

        self.account = account
        self._client = app.get_client(account)

        self._action = action
        self._exchange_list = exchange_list
        self._jid_from = jid_from

        self._ui = get_builder("roster_item_exchange.ui")
        self.set_child(self._ui.roster_item_exchange)

        self._connect(self._ui.cancel_button, "clicked", self._on_cancel_button_clicked)
        self._connect(self._ui.accept_button, "clicked", self._on_accept_button_clicked)

        contact = self._client.get_module("Contacts").get_contact(jid_from.bare)
        assert isinstance(
            contact, BareContact | GroupchatContact | GroupchatParticipant
        )

        # Set label depending on action
        type_label = ""
        if action == "add":
            type_label = _(
                "%(name)s (%(jid)s) would like to add some "
                "contacts to your contact list."
            ) % {"name": contact.name, "jid": self._jid_from.bare}
        elif action == "modify":
            type_label = _(
                "%(name)s (%(jid)s) would like to modify some "
                "contacts in your contact list."
            ) % {"name": contact.name, "jid": self._jid_from.bare}
        elif action == "delete":
            type_label = _(
                "%(name)s (%(jid)s) would like to delete some "
                "contacts from your contact list."
            ) % {"name": contact.name, "jid": self._jid_from.bare}
        self._ui.type_label.set_text(type_label)

        if message_body:
            buffer_ = self._ui.body_textview.get_buffer()
            buffer_.set_text(message_body)
        else:
            self._ui.body_scrolledwindow.set_visible(False)

        # Treeview
        self._model = Gtk.ListStore(bool, str, str, str, str)
        self._ui.items_list_treeview.set_model(self._model)
        # Columns
        renderer1 = Gtk.CellRendererToggle()
        renderer1.set_property("activatable", True)
        self._connect(renderer1, "toggled", self._on_toggled)
        title = ""
        if self._action == "add":
            title = _("Add")
        elif self._action == "modify":
            title = _("Modify")
        elif self._action == "delete":
            title = _("Delete")
        self._ui.items_list_treeview.insert_column_with_attributes(
            -1, title, renderer1, active=0
        )
        renderer2 = Gtk.CellRendererText()
        self._ui.items_list_treeview.insert_column_with_attributes(
            -1, _("JID"), renderer2, text=1
        )
        renderer3 = Gtk.CellRendererText()
        self._ui.items_list_treeview.insert_column_with_attributes(
            -1, _("Name"), renderer3, text=2
        )
        renderer4 = Gtk.CellRendererText()
        self._ui.items_list_treeview.insert_column_with_attributes(
            -1, _("Groups"), renderer4, text=3
        )

        if action == "add":
            self._add()
        elif action == "modify":
            self._modify()
        elif action == "delete":
            self._delete()

    def _cleanup(self) -> None:
        pass

    def _on_toggled(self, cell: Gtk.CellRendererToggle, path: str) -> None:
        iter_ = self._model.get_iter(path)
        self._model[iter_][0] = not cell.get_active()

    def _add(self) -> None:
        for jid in self._exchange_list:
            contact = self._client.get_module("Contacts").get_contact(jid)
            assert isinstance(contact, BareContact)
            name = self._exchange_list[jid][0]
            groups = ", ".join(self._exchange_list[jid][1])
            if not contact.is_in_roster:
                iter_ = self._model.append()
                self._model.set(iter_, {0: True, 1: jid, 2: name, 3: groups})

        self._ui.accept_button.set_label(_("Add"))

    def _modify(self) -> None:
        for jid in self._exchange_list:
            is_right = True
            contact = self._client.get_module("Contacts").get_contact(jid)
            assert isinstance(contact, BareContact)
            name = self._exchange_list[jid][0]
            groups = ", ".join(self._exchange_list[jid][1])

            if name != contact.name:
                is_right = False
            for group in self._exchange_list[jid][1]:
                if group not in contact.groups:
                    is_right = False
            if not is_right and contact.is_in_roster:
                iter_ = self._model.append()
                self._model.set(iter_, {0: True, 1: jid, 2: name, 3: groups})

        self._ui.accept_button.set_label(_("Modify"))

    def _delete(self) -> None:
        for jid in self._exchange_list:
            contact = self._client.get_module("Contacts").get_contact(jid)
            assert isinstance(contact, BareContact)
            name = self._exchange_list[jid][0]
            groups = ", ".join(self._exchange_list[jid][1])
            if contact.is_in_roster:
                iter_ = self._model.append()
                self._model.set(iter_, {0: True, 1: jid, 2: name, 3: groups})

        self._ui.accept_button.set_label(_("Delete"))

    def _on_accept_button_clicked(self, _button: Gtk.Button) -> None:
        iter_ = self._model.get_iter_first()
        if self._action == "add":
            count = 0
            while iter_:
                if self._model[iter_][0]:
                    count += 1
                    # It is selected
                    contact = self._client.get_module("Contacts").get_contact(
                        self._jid_from.bare
                    )
                    assert isinstance(
                        contact, BareContact | GroupchatContact | GroupchatParticipant
                    )
                    message = _(
                        "%(name)s %(jid)s suggested me to add you to my contact list."
                    ) % {"name": contact.name, "jid": self._jid_from.bare}
                    # Keep same groups and same nickname
                    groups: list[str] = []
                    groups = self._model[iter_][3].split(", ")
                    if groups == [""]:
                        groups = []
                    jid = self._model[iter_][1]
                    if app.jid_is_transport(str(self._jid_from)):
                        self._client.get_module("Presence").automatically_added.add(
                            JID.from_string(jid)
                        )

                    self._client.get_module("Presence").subscribe(
                        JID.from_string(jid),
                        msg=message,
                        name=self._model[iter_][2],
                        groups=groups,
                        auto_auth=True,
                    )
                iter_ = self._model.iter_next(iter_)
            InformationAlertDialog(
                _("Successfully Added Contacts"),
                i18n.ngettext(
                    "Added %(count)s contact", "Added %(count)s contacts", count
                )
                % {"count": count},
            )
        elif self._action == "modify":
            count = 0
            while iter_:
                if self._model[iter_][0]:
                    count += 1
                    # It is selected
                    jid = self._model[iter_][1]
                    # Keep same groups and same nickname
                    groups = self._model[iter_][3].split(", ")
                    if groups == [""]:
                        groups = []
                    self._client.get_module("Roster").set_item(
                        jid, self._model[iter_][2], groups
                    )

                iter_ = self._model.iter_next(iter_)
        elif self._action == "delete":
            count = 0
            while iter_:
                if self._model[iter_][0]:
                    count += 1
                    # It is selected
                    jid = self._model[iter_][1]
                    self._client.get_module("Presence").unsubscribe(
                        JID.from_string(jid)
                    )
                    self._client.get_module("Roster").delete_item(jid)
                iter_ = self._model.iter_next(iter_)
            InformationAlertDialog(
                _("Successfully Removed Contacts"),
                i18n.ngettext(
                    "Removed %(count)s contact", "Removed %(count)s contacts", count
                )
                % {"count": count},
            )
        self.close()

    def _on_cancel_button_clicked(self, _button: Gtk.Button) -> None:
        self.close()



# Copyright (C) 2005 Sebastian Estienne
# Copyright (C) 2005-2006 Andrew Sayman <lorien420 AT myrealbox.com>
# Copyright (C) 2005-2007 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2005-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2006 Travis Shirk <travis AT pobox.com>
# Copyright (C) 2006-2008 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2007 Julien Pivotto <roidelapluie AT gmail.com>
#                    Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2008 Brendan Taylor <whateley AT gmail.com>
#                    Jonathan Schleifer <js-gajim AT webkeks.org>
# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import hashlib
import logging
import platform
import sys
from pathlib import Path

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import configpaths
from gajim.common import events
from gajim.common import ged
from gajim.common.client import Client
from gajim.common.const import AvatarSize
from gajim.common.const import SimpleClientState
from gajim.common.ged import EventHelper
from gajim.common.helpers import allow_showing_notification
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import ResourceContact

from gajim.gtk import sound
from gajim.gtk.avatar import merge_avatars
from gajim.gtk.structs import AccountJidParam
from gajim.gtk.structs import OpenEventActionParams

MIN_WINDOWS_TOASTS_WIN_VERSION = 10240

if (
    sys.platform == "win32"
    and int(platform.version().split(".")[2]) >= MIN_WINDOWS_TOASTS_WIN_VERSION
) or TYPE_CHECKING:
    # Importing windows_toasts on an unsupported OS will throw an Exception
    from windows_toasts import InteractableWindowsToaster
    from windows_toasts import Toast
    from windows_toasts import ToastActivatedEventArgs
    from windows_toasts import ToastButton
    from windows_toasts import ToastDisplayImage
    from windows_toasts import ToastImage
    from windows_toasts import ToastImagePosition

log = logging.getLogger("gajim.gtk.notification")

NOTIFICATION_ICONS: dict[str, str] = {
    "muc-invitation": "gajim-group-chat-invitation",
    "subscription-request": "gajim-subscription-request",
}


_notification_backend = None


class NotificationBackend(EventHelper):
    def __init__(self) -> None:
        EventHelper.__init__(self)

        self.register_events(
            [
                ("notification", ged.GUI2, self._on_notification),
                ("notification-withdrawn", ged.GUI2, self._on_notification_withdrawn),
                ("account-enabled", ged.GUI2, self._on_account_enabled),
                ("chat-read", ged.GUI2, self._on_chat_read),
            ]
        )

        for client in app.get_clients():
            client.connect_signal("state-changed", self._on_client_state_changed)

    def _on_notification(self, event: events.Notification) -> None:
        if event.account and event.jid:
            client = app.get_client(event.account)
            contact = client.get_module("Contacts").get_contact(event.jid)

            if contact.is_muted:
                log.debug("Notifications muted for %s", contact)
                return

            if isinstance(contact, GroupchatContact) and event.resource is not None:
                participant = contact.get_resource(event.resource)
                if participant.is_blocked:
                    log.debug("Notifications blocked for %s", participant)
                    return

        if event.sound is not None:
            sound.play(event.sound, event.account)

        if not allow_showing_notification(event.account):
            return

        self._send(event)

    def _on_notification_withdrawn(self, event: events.NotificationWithdrawn) -> None:
        self._withdraw([event.context_id])

    def _on_account_enabled(self, event: events.AccountEnabled) -> None:
        client = app.get_client(event.account)
        client.connect_signal("state-changed", self._on_client_state_changed)

    def _on_chat_read(self, event: events.ChatRead) -> None:
        self._withdraw(["new-message", event.account, event.jid])

    def _on_client_state_changed(
        self, client: Client, _signal_name: str, state: SimpleClientState
    ) -> None:
        if not state.is_connected:
            return
        self._withdraw(["connection-failed", client.account])
        self._withdraw(["server-shutdown", client.account])

    def _send(self, event: events.Notification) -> None:
        raise NotImplementedError

    def _withdraw(self, details: list[Any]) -> None:
        raise NotImplementedError


class DummyBackend(NotificationBackend):
    def _send(self, event: events.Notification) -> None:
        pass

    def _withdraw(self, details: list[Any]) -> None:
        pass


class WindowsToastNotification(NotificationBackend):
    def __init__(self):
        NotificationBackend.__init__(self)

        if app.is_ms_store():
            from winrt.windows.applicationmodel import AppInfo

            assert AppInfo.current is not None
            aumid = AppInfo.current.app_user_model_id
        else:
            # Non MS Store version has to register an AUMID manually
            aumid = "Gajim.ToastNotification"
            self._register_notifier_aumid(aumid)

        self._toaster = InteractableWindowsToaster(
            applicationText="Gajim", notifierAUMID=aumid
        )

    def _register_notifier_aumid(self, aumid: str) -> None:
        """Register an AUMID for Gajim's toast notifications.
        This allows notifications issued by Gajim to have the right icon and title.
        Code taken from: https://github.com/DatGuy1/Windows-Toasts/blob/main/scripts/register_hkey_aumid.py
        """
        key_path = f"SOFTWARE\\Classes\\AppUserModelId\\{aumid}"

        image_path = (
            Path(sys.executable).parent.parent
            / "share"
            / "icons"
            / "hicolor"
            / "96x96"
            / "apps"
            / "gajim.png"
        )

        import winreg

        winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER)
        with winreg.CreateKeyEx(winreg.HKEY_CURRENT_USER, key_path) as master_key:
            winreg.SetValueEx(master_key, "DisplayName", 0, winreg.REG_SZ, "Gajim")
            winreg.SetValueEx(
                master_key, "IconUri", 0, winreg.REG_SZ, str(image_path.resolve())
            )

    def _send(self, event: events.Notification) -> None:
        toast = Toast()
        toast.text_fields = [event.title, event.text]

        toast_image = self._get_toast_image(event)
        if toast_image is not None:
            toast_display_image = ToastDisplayImage(
                image=toast_image,
                position=ToastImagePosition.AppLogo,
            )
            toast.AddImage(toast_display_image)

        for button in self._get_toast_buttons(event):
            toast.AddAction(button)

        toast.on_activated = self._on_activated

        self._toaster.show_toast(toast)

    def _withdraw(self, details: list[Any]) -> None:
        self._toaster.clear_toasts()

    def _on_activated(self, event: ToastActivatedEventArgs) -> None:
        # Calls need to be executed with GLib.idle_add to avoid threading issues,
        # because Toasts run in a different thread.
        if event.arguments is None:
            GLib.idle_add(app.window.present)
            return

        if event.arguments.startswith("open-event-"):
            serialized_data = event.arguments.split("open-event-")[1]
            params = OpenEventActionParams.from_serialized_string(
                serialized_data, GLib.VariantType("a{sv}")
            )
            if params is not None:
                GLib.idle_add(
                    app.app.activate_action,
                    f"{params.account}-open-event",
                    params.to_variant(),
                )
                return

        elif event.arguments.startswith("mark-as-read-"):
            serialized_data = event.arguments.split("mark-as-read-")[1]
            params = AccountJidParam.from_serialized_string(
                serialized_data, GLib.VariantType("a{sv}")
            )
            if params is not None:
                GLib.idle_add(
                    app.app.activate_action,
                    f"{params.account}-mark-as-read",
                    params.to_variant(),
                )
                return

        GLib.idle_add(app.window.present)

    def _get_toast_image(self, event: events.Notification) -> ToastImage | None:
        if event.type == "incoming-message":
            assert event.jid is not None
            texture = _get_avatar_texture_for_notification(
                event.account, event.jid, event.resource
            )
            return ToastImage(_get_path_for_avatar_texture(texture))

        icon_name = event.sub_type or event.type
        icon_name = NOTIFICATION_ICONS.get(icon_name)
        if icon_name is None:
            return None

        return ToastImage(_get_icon_path(icon_name, 32))

    def _get_toast_buttons(self, event: events.Notification) -> list[ToastButton]:
        toast_buttons: list[ToastButton] = []

        jid = "" if event.jid is None else str(event.jid)

        params = OpenEventActionParams(
            type=event.type,
            sub_type=event.sub_type or "",
            account=event.account,
            jid=jid,
            context_id=event.context_id,
        )

        button = ToastButton(
            content=_("Open"), arguments=f"open-event-{params.serialize()}"
        )
        toast_buttons.append(button)

        if event.type == "incoming-message":
            assert isinstance(event.jid, JID)
            params = AccountJidParam(account=event.account, jid=event.jid)

            button = ToastButton(
                content=_("Mark as Read"),
                arguments=f"mark-as-read-{params.serialize()}",
            )
            toast_buttons.append(button)

        return toast_buttons


class GioNotification(NotificationBackend):
    _action_types = [
        "connection-failed",
        "server-shutdown",
        "muc-invitation",
        "incoming-call",
        "incoming-message",
        "subscription-request",
        "reaction-updated",
    ]

    def __init__(self):
        NotificationBackend.__init__(self)

    def _send(self, event: events.Notification) -> None:
        notification = Gio.Notification()
        notification.set_title(event.title)

        text = event.text
        if "<img" in event.text:
            # Escape text with img tags
            # https://dev.gajim.org/gajim/gajim/-/issues/12349
            # https://gitlab.gnome.org/GNOME/glib/-/issues/3720
            text = GLib.markup_escape_text(event.text)

        notification.set_body(text)
        notification.set_priority(Gio.NotificationPriority.NORMAL)

        icon = self._make_icon(event)
        if icon is not None:
            notification.set_icon(icon)

        self._add_actions(event, notification)
        notification_id = self._make_notification_id(event)

        log.info("Sending notification: %s", notification_id)
        app.app.send_notification(notification_id, notification)

    def _add_actions(
        self, event: events.Notification, notification: Gio.Notification
    ) -> None:
        if event.type not in self._action_types:
            return

        jid = ""
        if event.jid is not None:
            jid = str(event.jid)

        params = OpenEventActionParams(
            type=event.type,
            sub_type=event.sub_type or "",
            account=event.account,
            jid=jid,
            context_id=event.context_id,
        )

        action = f"app.{event.account}-open-event"

        if sys.platform != "darwin":
            # The MacOS backend only supports one button, we can do without
            # the "Open" button
            notification.add_button_with_target(_("Open"), action, params.to_variant())

        notification.set_default_action_and_target(action, params.to_variant())

        if event.type == "incoming-message":
            action = f"app.{event.account}-mark-as-read"
            params = AccountJidParam(account=event.account, jid=JID.from_string(jid))
            notification.add_button_with_target(
                _("Mark as Read"), action, params.to_variant()
            )

    def _make_notification_id(self, event: events.Notification) -> str | None:
        if event.context_id:
            return event.context_id

        if event.type in ("connection-failed", "server-shutdown"):
            return self._make_id([event.type, event.account])

        if event.type == "incoming-message":
            return self._make_id(["new-message", event.account, str(event.jid)])

        return None

    def _make_icon(self, event: events.Notification) -> Gio.Icon | None:
        if event.type == "incoming-message":
            assert event.jid is not None

            if app.is_flatpak():
                return _get_avatar_bytes_icon(event.account, event.jid, event.resource)

            return _get_avatar_file_icon(event.account, event.jid, event.resource)

        icon_name = event.sub_type or event.type
        icon_name = NOTIFICATION_ICONS.get(icon_name)
        if icon_name is None:
            return None

        return _get_file_icon(icon_name, 96)

    def _withdraw(self, details: list[Any]) -> None:
        notification_id = self._make_id(details)

        log.info("Withdraw notification: %s", notification_id)
        app.app.withdraw_notification(notification_id)

    @staticmethod
    def _make_id(details: list[Any]) -> str:
        return ",".join(map(str, details))


def _get_participant_texture(
    gc_contact: GroupchatContact, resource: str
) -> Gdk.Texture:
    size = AvatarSize.NOTIFICATION
    muc_avatar = gc_contact.get_avatar(size, 1)
    participant_avatar = gc_contact.get_resource(resource).get_avatar(
        size, 1, add_show=False
    )

    return merge_avatars(muc_avatar, participant_avatar)


def _get_avatar_texture_for_notification(
    account: str, jid: JID | str, resource: str | None
) -> Gdk.Texture:
    size = AvatarSize.NOTIFICATION
    client = app.get_client(account)
    contact = client.get_module("Contacts").get_contact(jid)
    if isinstance(contact, GroupchatContact):
        if resource:
            return _get_participant_texture(contact, resource)
        else:
            return contact.get_avatar(size, 1)

    assert not isinstance(contact, ResourceContact)
    return contact.get_avatar(size, 1, add_show=False)


def _get_file_icon(name: str, size: int) -> Gio.FileIcon:
    path = _get_icon_path(name, size)
    return Gio.FileIcon(file=Gio.File.new_for_path(str(path)))


def _get_icon_path(name: str, size: int) -> Path:
    return (
        configpaths.get("ICONS")
        / "hicolor"
        / f"{size}x{size}"
        / "status"
        / f"{name}.png"
    )


def _get_avatar_bytes_icon(
    account: str, jid: JID | str, resource: str | None
) -> Gio.BytesIcon:
    texture = _get_avatar_texture_for_notification(account, jid, resource)
    png_bytes = texture.save_to_png_bytes()
    return Gio.BytesIcon(bytes=png_bytes)


def _get_avatar_file_icon(
    account: str, jid: JID | str, resource: str | None
) -> Gio.FileIcon:
    texture = _get_avatar_texture_for_notification(account, jid, resource)
    path = _get_path_for_avatar_texture(texture)
    return Gio.FileIcon(file=Gio.File.new_for_path(str(path)))


def _get_path_for_avatar_texture(texture: Gdk.Texture) -> Path:
    png_bytes = texture.save_to_png_bytes()
    png_bytes_data = png_bytes.get_data()
    assert png_bytes_data is not None
    path = configpaths.get("AVATAR_ICONS") / hashlib.sha1(png_bytes_data).hexdigest()
    if not path.exists():
        texture.save_to_png(str(path))
    return path


def get_notification_backend() -> NotificationBackend:
    if sys.platform == "win32":
        if int(platform.version().split(".")[2]) >= MIN_WINDOWS_TOASTS_WIN_VERSION:
            try:
                return WindowsToastNotification()
            except OSError as e:
                log.error(
                    "Error while trying to initialize notification backend: %s", e
                )
        return DummyBackend()

    return GioNotification()


def init() -> None:
    global _notification_backend  # pylint: disable=global-statement
    _notification_backend = get_notification_backend()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging
import textwrap
from datetime import datetime
from datetime import timedelta

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import Gtk
from nbxmpp.namespaces import Namespace

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.const import Trust
from gajim.common.const import TRUST_SYMBOL_DATA
from gajim.common.i18n import _
from gajim.common.i18n import is_rtl_text
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.open_graph_parser import OpenGraphData
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageState
from gajim.common.storage.archive.const import MessageType
from gajim.common.storage.archive.models import Message
from gajim.common.types import ChatContactT
from gajim.common.util.muc import message_needs_highlight
from gajim.common.util.preview import GeoPreview
from gajim.common.util.preview import get_preview_data
from gajim.common.util.preview import UrlPreview
from gajim.common.util.text import format_fingerprint
from gajim.common.util.user_strings import get_moderation_text
from gajim.common.util.user_strings import get_retraction_text

from gajim.gtk.conversation.message_widget import MessageWidget
from gajim.gtk.conversation.reactions_bar import ReactionsBar
from gajim.gtk.conversation.rows.base import BaseRow
from gajim.gtk.conversation.rows.widgets import AvatarBox
from gajim.gtk.conversation.rows.widgets import DateTimeLabel
from gajim.gtk.conversation.rows.widgets import MentionIndicator
from gajim.gtk.conversation.rows.widgets import MessageIcons
from gajim.gtk.conversation.rows.widgets import NicknameLabel
from gajim.gtk.menus import GajimMenu
from gajim.gtk.menus import get_chat_row_menu
from gajim.gtk.preview.geo import GeoPreviewWidget
from gajim.gtk.preview.open_graph import OpenGraphPreviewWidget
from gajim.gtk.preview.preview import PreviewWidget
from gajim.gtk.referenced_message import ReferencedMessageNotFoundWidget
from gajim.gtk.referenced_message import ReferencedMessageWidget
from gajim.gtk.util.misc import check_finalize
from gajim.gtk.util.misc import container_remove_all
from gajim.gtk.util.misc import get_avatar_for_message
from gajim.gtk.util.misc import get_contact_name_for_message

log = logging.getLogger("gajim.gtk.conversation.rows.message")

MERGE_TIMEFRAME = timedelta(seconds=120)


class MessageRow(BaseRow):
    def __init__(self, contact: ChatContactT, message: Message) -> None:
        BaseRow.__init__(self, contact.account)
        self.set_selectable(True)
        self.type = "chat"
        self._contact = contact
        self._message = message

        self.timestamp = message.timestamp.astimezone()
        self.db_timestamp = message.timestamp.timestamp()
        self.direction = ChatDirection(message.direction)
        self._is_outgoing = self.direction == ChatDirection.OUTGOING

        # Classes for identifying message direction via CSS
        if self._is_outgoing:
            self.add_css_class("outgoing-message")
        else:
            self.add_css_class("incoming-message")

        self.orig_pk = message.pk
        self.pk = self.orig_pk

        self.encryption = None
        self.securitylabel = None

        assert message.text is not None
        self._original_text = message.text
        self._original_message = message

        self._is_retracted = bool(message.moderation or message.retraction)
        self._is_blocked = False
        self._has_receipt = False

        self._avatar_box = AvatarBox(contact)

        self._mention_indicator = MentionIndicator()

        self._ref_message_widget = None

        self._meta_box = Gtk.Box(spacing=6)
        self._meta_box.set_hexpand(True)

        self._bottom_box = Gtk.Box(spacing=6)

        self.grid.attach(self._avatar_box, 0, 0, 1, 2)
        self.grid.attach(self._mention_indicator, 1, 0, 1, 2)
        self.grid.attach(self._meta_box, 2, 0, 1, 1)
        self.grid.attach(self._bottom_box, 2, 1, 1, 1)

        self._reactions_bar = ReactionsBar(self, self._contact)
        self.grid.attach(self._reactions_bar, 2, 2, 1, 1)

        self._redraw_content()

    @classmethod
    def from_db_row(cls, contact: ChatContactT, message: Message) -> MessageRow:
        return cls(contact, message)

    @property
    def state(self) -> int:
        return self._message.state

    @property
    def has_receipt(self) -> bool:
        return self._has_receipt

    @property
    def is_merged(self) -> bool:
        return self._merged

    @property
    def has_mention(self) -> bool:
        return self._mention_indicator.get_visible()

    @property
    def occupant_id(self) -> str | None:
        if self._message.occupant is None:
            return None
        return self._message.occupant.id

    def _redraw_content(self) -> None:
        self.set_merged(False)
        self.remove_css_class("retracted-message")

        check_finalize(self._meta_box, only_children=True)
        check_finalize(self._bottom_box, only_children=True)
        container_remove_all(self._meta_box)
        container_remove_all(self._bottom_box)

        message = self._original_message

        self.name = get_contact_name_for_message(message, self._contact)

        avatar = get_avatar_for_message(
            message, self._contact, self.get_scale_factor(), AvatarSize.ROSTER
        )
        self._avatar_box.set_from_paintable(avatar)
        self._avatar_box.set_name(self.name)

        self._meta_box.append(NicknameLabel(self.name, self._is_outgoing))
        self._meta_box.append(DateTimeLabel(self.timestamp))

        self._message_icons = MessageIcons()
        self._meta_box.append(self._message_icons)

        self._is_blocked = False
        if message.occupant is not None and message.occupant.blocked:
            self._set_blocked()
            return

        if message.is_retracted():
            self._set_retracted(message)
            return

        # From here on, if this is a correction all data must
        # be taken from the correction
        if corrected_message := message.get_last_correction():
            message = corrected_message
            self._message = corrected_message

        self.pk = message.pk

        self.encryption = message.encryption
        self.securitylabel = message.security_label

        assert message.text is not None
        self.text = message.text

        match preview := get_preview_data(self.text, message.oob):
            case UrlPreview():
                self._message_widget = PreviewWidget(
                    self._contact.account, preview, self._is_outgoing, self._muc_context
                )

            case GeoPreview():
                self._message_widget = GeoPreviewWidget(preview)

            case _:
                self._message_widget = MessageWidget(self._contact.account)
                self._message_widget.add_with_styling(self.text, nickname=self.name)
                if self._contact.is_groupchat and not self._is_outgoing:
                    self._apply_highlight(self.text)

        if message.reply is not None:
            referenced_message = message.get_referenced_message()
            if referenced_message is None:
                self._ref_message_widget = ReferencedMessageNotFoundWidget()
            else:
                self._ref_message_widget = ReferencedMessageWidget(
                    self._contact, referenced_message
                )

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=3)
        if self._ref_message_widget is not None:
            box.append(self._ref_message_widget)

        box.append(self._message_widget)

        if message.og and self._contact.settings.get("enable_link_preview"):
            og_box = Gtk.Box(
                orientation=Gtk.Orientation.VERTICAL,
                margin_top=6,
                spacing=6,
            )
            og_clamp = Adw.Clamp(
                child=og_box,
                maximum_size=500,
                tightening_threshold=600,
                halign=Gtk.Align.START,
            )
            for og in message.og:
                og_data = OpenGraphData.from_model(og)
                og_box.append(
                    OpenGraphPreviewWidget(og.about, og_data=og_data, pk=og.pk)
                )

            box.append(og_clamp)

        self._bottom_box.append(box)

        self._set_text_direction(self.text)

        if corrected_message:
            self._set_correction()

        self._reactions_bar.update_from_reactions(
            self._original_message.get_reactions()
        )

        encryption_data = self._get_encryption_data(message.encryption)
        if encryption_data is not None:
            self._message_icons.set_encrytion_icon_data(*encryption_data)
            self._message_icons.set_encryption_icon_visible(True)

        sec_label_data = self._get_security_labels_data(message.security_label)
        if sec_label_data is not None:
            self._message_icons.set_security_label_data(*sec_label_data)
            self._message_icons.set_security_label_visible(True)

        self._set_receipt(message.receipt is not None)

        if self._contact.is_groupchat and self.direction == ChatDirection.OUTGOING:
            self._message_icons.set_message_state_icon(MessageState(message.state))

        if message.error is not None:
            if message.error.text is not None:
                error_text = f"{message.error.text} ({message.error.condition})"
            else:
                error_text = message.error.condition
            self._set_error(error_text)

    def _set_text_direction(self, text: str) -> None:
        global_rtl = self._message_widget.get_direction() == Gtk.TextDirection.RTL
        message_rtl = is_rtl_text(text)

        if (global_rtl and not message_rtl) or (not global_rtl and message_rtl):
            # LTR message in an RTL environment (align to the left) or
            # RTL message in an LTR environment (align to the right)
            self._bottom_box.set_halign(Gtk.Align.END)

    @property
    def _muc_context(self) -> str | None:
        if isinstance(self._contact, GroupchatContact | GroupchatParticipant):
            return self._contact.muc_context
        return None

    def get_chat_row_menu(
        self,
    ) -> GajimMenu:
        copy_text = self._get_copy_text(self.get_text(), self.name, self.timestamp)

        return get_chat_row_menu(
            contact=self._contact,
            copy_text=copy_text,
            message=self._message,
            original_message=self._original_message,
        )

    @staticmethod
    def _get_copy_text(text: str, name: str, timestamp: datetime) -> str | None:
        # Text can be an empty string
        # e.g. if a preview has not been loaded yet
        if not text:
            return None

        timestamp_formatted = timestamp.strftime(app.settings.get("date_time_format"))

        copy_text = f"{timestamp_formatted} - {name}: "
        if text.startswith(("```", "> ")):
            # Prepend a line break in order to keep code block/quotes rendering
            copy_text += "\n"
        copy_text += text
        return copy_text

    def enable_selection_mode(self) -> None:
        if isinstance(self._message_widget, MessageWidget):
            self._message_widget.set_selectable(False)

    def disable_selection_mode(self) -> None:
        if isinstance(self._message_widget, MessageWidget):
            self._message_widget.set_selectable(True)

    def _get_security_labels_data(
        self, security_labels: mod.SecurityLabel | None
    ) -> tuple[str, str] | None:
        if security_labels is None:
            return None

        if not app.settings.get_account_setting(
            self._account, "enable_security_labels"
        ):
            return None

        displaymarking = GLib.markup_escape_text(security_labels.displaymarking)
        if displaymarking:
            bgcolor = security_labels.bgcolor
            fgcolor = security_labels.fgcolor
            markup = (
                f'<span size="small" bgcolor="{bgcolor}" '
                f'fgcolor="{fgcolor}"><tt>{displaymarking}</tt></span>'
            )
            return displaymarking, markup
        return None

    def _apply_highlight(self, text: str) -> None:
        assert isinstance(self._contact, GroupchatContact)
        if self._contact.nickname is None:
            return

        needs_highlight = message_needs_highlight(
            text, self._contact.nickname, self._client.get_own_jid().bare
        )
        self._mention_indicator.set_visible(needs_highlight)
        self._message_icons.set_mention_icon_visibe(needs_highlight)

    def is_same_sender(self, message: MessageRow) -> bool:
        return message.name == self.name

    def is_same_encryption(self, message: MessageRow) -> bool:
        c_enc = message.encryption
        o_enc = self.encryption
        if c_enc is None and o_enc is None:
            return True

        if c_enc is not None and o_enc is not None:
            if c_enc.protocol == o_enc.protocol and c_enc.trust == o_enc.trust:
                return True

        return False

    def is_same_securitylabels(self, message: MessageRow) -> bool:
        sec1 = message.securitylabel
        sec2 = self.securitylabel

        if sec1 is None and sec2 is None:
            return True

        if sec1 is None or sec2 is None:
            return False

        return sec1.label_hash == sec2.label_hash

    def is_same_state(self, message: MessageRow) -> bool:
        return message.state == self._message.state

    def has_same_receipt_status(self, message: MessageRow) -> bool:
        if not app.settings.get("positive_184_ack"):
            return True
        return self._has_receipt == message.has_receipt

    def is_same_mention_state(self, message: MessageRow) -> bool:
        return self.has_mention is message.has_mention

    def is_mergeable(self, message: MessageRow) -> bool:
        if message.type != self.type:
            return False
        if not self.is_same_state(message):
            return False
        if message.direction != self.direction:
            return False
        if self._message.correction_id is not None:
            return False
        if not self.is_same_sender(message):
            return False
        if not self.is_same_encryption(message):
            return False
        if not self.is_same_securitylabels(message):
            return False
        if not self.has_same_receipt_status(message):
            return False
        if not self.is_same_mention_state(message):
            return False
        return abs(message.timestamp - self.timestamp) < MERGE_TIMEFRAME

    def get_text(self) -> str:
        return self._message_widget.get_text()

    def _get_encryption_data(
        self,
        encryption_data: mod.Encryption | None,
    ) -> tuple[str, str, str] | None:
        contact_encryption = self._contact.settings.get("encryption")
        if encryption_data is None:
            if not contact_encryption:
                return None

            icon = "channel-insecure-symbolic"
            color = "unencrypted-color"
            tooltip = _("Not encrypted")
        else:
            tooltip = _("Encrypted (%s)") % (encryption_data.protocol)
            icon, trust_tooltip, color = TRUST_SYMBOL_DATA[Trust(encryption_data.trust)]
            tooltip = f"{tooltip}\n{trust_tooltip}"
            if encryption_data.key != "Unknown":
                fingerprint = format_fingerprint(
                    encryption_data.key, encryption_data.protocol, wrap=True
                )
                tooltip = f"{tooltip}\n<tt>{fingerprint}</tt>"

        return icon, color, tooltip

    def set_acknowledged(self, pk: int) -> None:
        if self._original_message.pk == pk:
            message = self._original_message
        elif self._message.pk == pk:
            message = self._message
        else:
            raise ValueError("Acknowledged unknown message")

        app.storage.archive.refresh(message, ["stanza_id", "state"])

        if self._is_retracted:
            return

        self._message_icons.set_message_state_icon(MessageState(message.state))

    def set_blocked(self, blocked: bool) -> None:
        if blocked != self._is_blocked:
            app.storage.archive.refresh(self._original_message, ["occupant"])
            self._redraw_content()

    def _set_receipt(self, value: bool) -> None:
        if self._is_retracted:
            return

        self._has_receipt = value
        self._message_icons.set_receipt_icon_visible(value)

    def set_receipt(self, receipt_id: str) -> None:
        if self._message.id == receipt_id:
            self._set_receipt(True)

    def _set_error(self, tooltip: str) -> None:
        if self._is_retracted:
            return

        self._message_icons.hide_message_state_icon()
        self._message_icons.set_error_icon_visible(True)
        self._message_icons.set_error_tooltip(tooltip)

    def set_error(self, error_id: str, text: str) -> None:
        if self._message.id == error_id:
            self._set_error(text)
            self.set_merged(False)

    def update_reactions(self) -> None:
        app.storage.archive.refresh(
            self._original_message, ["corrections", "reactions"]
        )
        if self._is_retracted:
            return

        self._reactions_bar.update_from_reactions(
            self._original_message.get_reactions()
        )

    def update_retractions(self) -> None:
        app.storage.archive.refresh(
            self._original_message, ["corrections", "retraction", "moderation"]
        )
        self._redraw_content()

    def update_corrections(self) -> None:
        app.storage.archive.refresh(self._original_message, ["corrections"])
        self._redraw_content()

    def send_reaction(self, emoji: str, toggle: bool = True) -> None:
        """Adds or removes 'emoji' from this message's reactions and sends the result.

        Args:
          emoji: Reaction emoji to add or remove
          toggle: Whether an existing emoji should be removed from the set
        """
        reaction_id = self._original_message.id
        if self._original_message.type == MessageType.GROUPCHAT:
            reaction_id = self._original_message.stanza_id

        if reaction_id is None:
            log.warning("No reaction id")
            return

        our_reactions = self._reactions_bar.get_our_reactions()
        if emoji in our_reactions and not toggle:
            log.info("Not toggling reaction <%s>", emoji)
            return

        if emoji in our_reactions:
            our_reactions.discard(emoji)
        elif self._contact.reactions_per_user == 1:
            our_reactions = {emoji}
        else:
            our_reactions.add(emoji)

        client = app.get_client(self._contact.account)
        client.get_module("Reactions").send_reaction(
            contact=self._contact, reaction_id=reaction_id, reactions=our_reactions
        )

    def _set_retracted(self, message: Message) -> None:
        if message.moderation is not None:
            text = get_moderation_text(message.moderation.by, message.moderation.reason)

        elif message.retraction is not None:
            text = get_retraction_text(message.retraction.timestamp)

        else:
            raise ValueError("Unable to determine retraction text")

        self._set_disabled(text)
        self._is_retracted = True

    def _set_blocked(self) -> None:
        self._set_disabled(_("Messages from this users are blocked"))
        self._is_blocked = True

    def _set_disabled(self, text: str) -> None:
        self.text = text

        self._message_widget = MessageWidget(self._account)
        self._message_widget.add_with_styling(text)
        self._set_text_direction(text)
        self._bottom_box.append(self._message_widget)

        self._reactions_bar.set_visible(False)

        self.add_css_class("retracted-message")

    def _set_correction(self) -> None:
        original_text = textwrap.fill(
            self._original_text, width=150, max_lines=10, placeholder="…"
        )
        self._message_icons.set_correction_tooltip(
            _("Message corrected. Original message:\n%s") % original_text
        )
        self._message_icons.set_correction_icon_visible(True)

    def update_avatar(self) -> None:
        avatar = get_avatar_for_message(
            self._original_message,
            self._contact,
            self.get_scale_factor(),
            AvatarSize.ROSTER,
        )
        self._avatar_box.set_from_paintable(avatar)

    def set_merged(self, merged: bool) -> None:
        self._merged = merged
        if merged:
            self.add_css_class("merged")
            self._meta_box.set_visible(False)
            self._meta_box.set_visible(False)
        else:
            self.remove_css_class("merged")
            self._meta_box.set_visible(True)

        self._avatar_box.set_merged(merged)

    def can_reply(self) -> bool:
        if self._is_retracted:
            return False

        if isinstance(self._contact, GroupchatContact):
            if self._message.stanza_id is None:
                return False

            if not self._contact.is_joined:
                return False

            self_contact = self._contact.get_self()
            assert self_contact is not None
            return not self_contact.role.is_visitor

        return self._original_message.id is not None

    def can_react(self) -> bool:
        if self._is_retracted:
            return False

        if not app.account_is_connected(self._contact.account):
            return False

        if isinstance(self._contact, GroupchatContact):
            if not self._contact.is_joined:
                return False

            self_contact = self._contact.get_self()
            assert self_contact is not None
            if self_contact.role.is_visitor:
                return False

            if self._message.stanza_id is None:
                return False

            if self._contact.muc_context == "public":
                return self._contact.supports(Namespace.OCCUPANT_ID)

            return True

        return self._original_message.id is not None



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import datetime as dt
import logging
from collections.abc import Iterable
from pathlib import Path

import pysequoia as pys
import sqlalchemy as sa
from nbxmpp.protocol import JID
from pysequoia import packet as pyspacket
from sqlalchemy.engine import Engine
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session

import gajim.common.storage.openpgp.models as mod
from gajim.common import app
from gajim.common import configpaths
from gajim.common.const import Trust
from gajim.common.const import VALUE_MISSING
from gajim.common.const import ValueMissingT
from gajim.common.storage.base import AlchemyStorage
from gajim.common.storage.base import timeit
from gajim.common.storage.base import with_session
from gajim.common.storage.openpgp.models import Account
from gajim.common.storage.openpgp.models import Base

CURRENT_USER_VERSION = 1

log = logging.getLogger("gajim.c.storage.openpgp")


class OpenPGPStorage(AlchemyStorage):
    def __init__(self, *, in_memory: bool = False, path: Path | None = None) -> None:
        if path is None:
            path = Path(configpaths.get("MY_DATA")) / "openpgp.db"

        AlchemyStorage.__init__(
            self,
            log,
            None if in_memory else path,
            pragma={
                "secure_delete": "on",
            },
        )

        self._account_pks: dict[str, int] = {}

    def _create_table(self, session: Session, engine: Engine) -> None:
        Base.metadata.create_all(engine)
        self.set_user_version(CURRENT_USER_VERSION)

    def _migrate(self) -> None:
        pass

    def _get_account_pk(self, session: Session, account: str) -> int:
        pk = self._account_pks.get(account)
        if pk is not None:
            return pk

        jid_str = app.get_jid_from_account(account)
        jid = JID.from_string(jid_str)

        pk = session.scalar(sa.select(Account.pk).where(Account.jid == jid))
        if pk is None:
            acc = Account(jid=jid)
            session.add(acc)
            session.flush()
            pk = acc.pk

        self._account_pks[account] = pk
        return pk

    @with_session
    @timeit
    def store_secret_key(
        self, session: Session, account: str, key: pys.Cert, backup_password: str | None
    ) -> None:
        assert key.secrets is not None

        fk_account_pk = self._get_account_pk(session, account)
        secret = mod.Secret(
            fk_account_pk=fk_account_pk,
            key=bytes(key.secrets),
            backup_password=backup_password,
        )
        log.info("Store secret key for %s", fk_account_pk)
        session.add(secret)

        try:
            session.commit()
        except Exception:
            log.exception("Unable to store secret key")

    @with_session
    @timeit
    def store_secret_key_backup_password(
        self, session: Session, account: str, backup_password: str
    ) -> None:
        log.info("Store secret key backup password for %s", account)

        fk_account_pk = self._get_account_pk(session, account)
        stmt = (
            sa.update(mod.Secret)
            .where(
                mod.Secret.fk_account_pk == fk_account_pk,
            )
            .values(backup_password=backup_password)
        )
        session.execute(stmt)

    @with_session
    @timeit
    def get_secret_key(
        self, session: Session, account: str
    ) -> tuple[mod.Secret, dt.datetime] | None:

        fk_account_pk = self._get_account_pk(session, account)
        stmt = sa.select(mod.Secret).where(mod.Secret.fk_account_pk == fk_account_pk)
        if row := session.scalar(stmt):
            date = None
            pile = pyspacket.PacketPile.from_bytes(row.key)
            for packet in pile:
                if packet.tag == pyspacket.Tag.SecretKey:
                    if date := packet.key_created:
                        break

            return row, date or dt.datetime(1990, 1, 1, tzinfo=dt.UTC)

        return None

    @with_session
    @timeit
    def store_public_key(
        self,
        session: Session,
        account: str,
        jid: JID,
        cert: pys.Cert,
        trust: Trust,
    ) -> None:

        fk_account_pk = self._get_account_pk(session, account)

        fingerprint = cert.fingerprint.upper()

        public = mod.Public(
            fk_account_pk=fk_account_pk,
            remote_jid=jid,
            key=cert,
            fingerprint=fingerprint,
            trust=trust,
        )
        log.info("Store public key for %s, %s, %s", jid, fingerprint, trust)
        session.add(public)

        try:
            session.commit()
        except IntegrityError:
            log.debug("Unable to store public key, key is already stored")

        except Exception:
            log.exception("Unable to store public key")

    @with_session
    @timeit
    def update_public_keys(
        self,
        session: Session,
        account: str,
        jid: JID,
        fingerprints: Iterable[str] | None = None,
        *,
        label: str | ValueMissingT | None = VALUE_MISSING,
        trust: Trust | ValueMissingT = VALUE_MISSING,
        active: bool | ValueMissingT = VALUE_MISSING,
    ) -> None:

        fk_account_pk = self._get_account_pk(session, account)

        values = {
            "label": label,
            "trust": trust,
            "active": active,
        }
        values = {k: v for k, v in values.items() if v is not VALUE_MISSING}

        fingerprints = set(map(str.upper, fingerprints or set()))

        log.info("Update public key for %s, %s, %s", jid, fingerprints, values)

        stmt = sa.update(mod.Public).where(
            mod.Public.fk_account_pk == fk_account_pk,
            mod.Public.remote_jid == jid,
        )

        if fingerprints:
            stmt = stmt.where(mod.Public.fingerprint.in_(fingerprints))

        stmt = stmt.values(**values)
        session.execute(stmt)

    @with_session
    @timeit
    def set_last_seen(
        self,
        session: Session,
        account: str,
        jid: JID,
        fingerprint: str,
        last_seen: dt.datetime,
    ) -> None:

        fk_account_pk = self._get_account_pk(session, account)

        log.info("Update last seen for %s, %s, %s", jid, fingerprint, last_seen)

        stmt = sa.update(mod.Public).where(
            mod.Public.fk_account_pk == fk_account_pk,
            mod.Public.remote_jid == jid,
            mod.Public.fingerprint == fingerprint.upper(),
            sa.or_(mod.Public.last_seen.is_(None), mod.Public.last_seen < last_seen),
        )

        stmt = stmt.values(last_seen=last_seen)
        session.execute(stmt)

    def get_public_key_from_secret(
        self, account: str, jid: JID, cert: pys.Cert
    ) -> mod.Public:

        return mod.Public(
            fk_account_pk=0,
            remote_jid=jid,
            key=cert,
            fingerprint=cert.fingerprint.upper(),
            trust=Trust.VERIFIED,
        )

    @with_session
    @timeit
    def get_public_key(
        self, session: Session, account: str, jid: JID, fingerprint: str
    ) -> mod.Public | None:

        fk_account_pk = self._get_account_pk(session, account)

        stmt = sa.select(mod.Public).where(
            mod.Public.fk_account_pk == fk_account_pk,
            mod.Public.remote_jid == jid,
            mod.Public.fingerprint == fingerprint.upper(),
        )
        return session.scalar(stmt)

    @with_session
    @timeit
    def get_public_keys(
        self,
        session: Session,
        account: str,
        jids: list[JID],
        *,
        trust: list[Trust] | None = None,
        only_active: bool = True,
    ) -> list[mod.Public]:

        fk_account_pk = self._get_account_pk(session, account)

        stmt = sa.select(mod.Public).where(
            mod.Public.fk_account_pk == fk_account_pk,
            mod.Public.remote_jid.in_(jids),
        )
        if trust is not None:
            stmt = stmt.where(mod.Public.trust.in_(trust))

        if only_active:
            stmt = stmt.where(mod.Public.active == only_active)

        return list(session.scalars(stmt).all())

    @with_session
    @timeit
    def remove_public_key(
        self, session: Session, account: str, jid: JID, fingerprint: str
    ) -> None:

        fk_account_pk = self._get_account_pk(session, account)

        stmt = sa.delete(mod.Public).where(
            mod.Public.fk_account_pk == fk_account_pk,
            mod.Public.remote_jid == jid,
            mod.Public.fingerprint == fingerprint.upper(),
        )
        session.execute(stmt)

    @timeit
    def get_known_fingerprints(
        self,
        account: str,
        jids: list[JID],
        *,
        trust: list[Trust] | None = None,
    ) -> set[str]:

        keys = self.get_public_keys(account, jids, trust=trust, only_active=False)
        return {k.fingerprint for k in keys}

    @with_session
    @timeit
    def remove_account(self, session: Session, account: str) -> None:
        fk_account_pk = self._get_account_pk(session, account)

        session.execute(sa.delete(Account).where(Account.pk == fk_account_pk))

        self._account_pks.pop(account)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging

from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app

from gajim.gtk.audio_player import AudioPreviewState
from gajim.gtk.preview.audio_visualizer import AudioVisualizerWidget
from gajim.gtk.util.classes import SignalManager

log = logging.getLogger("gajim.gtk.preview_audio_waveform_navigator")


class AudioWaveformNavigator(AudioVisualizerWidget, SignalManager):
    __gtype_name__ = "AudioWaveformNavigator"

    __gsignals__ = {
        "seeked": (
            GObject.SignalFlags.RUN_LAST,
            None,
            (float,),
        ),
    }

    def __init__(
        self, preview_state: AudioPreviewState, width: int = 340, height: int = 50
    ) -> None:
        AudioVisualizerWidget.__init__(
            self,
            width=width,
            height=height,
            is_seekable=True,
        )
        SignalManager.__init__(self)

        self._seek_ts = -1.0
        self._cursor_pos = 0.0
        self._user_holds_position_slider = False
        self._preview_state = preview_state

        self._scroll_step = 1e9  # 1 second

        self._is_ltr = bool(self.get_direction() == Gtk.TextDirection.LTR)

        gesture_seek_click = Gtk.GestureClick(
            button=Gdk.BUTTON_PRIMARY, propagation_phase=Gtk.PropagationPhase.CAPTURE
        )
        self._connect(gesture_seek_click, "pressed", self._on_button_pressed)
        self._connect(gesture_seek_click, "released", self._on_button_released)

        controller_motion = Gtk.EventControllerMotion(
            propagation_phase=Gtk.PropagationPhase.CAPTURE
        )
        self._connect(controller_motion, "motion", self._on_cursor_move)

        controller_scroll = Gtk.EventControllerScroll(
            flags=Gtk.EventControllerScrollFlags.HORIZONTAL,
            propagation_phase=Gtk.PropagationPhase.CAPTURE,
        )
        self._connect(controller_scroll, "scroll", self._on_scroll)

        self.add_controller(controller_scroll)
        self.add_controller(controller_motion)
        self.add_controller(gesture_seek_click)

    def run_destroy(self) -> None:
        self._disconnect_all()
        AudioVisualizerWidget.run_destroy(self)
        app.check_finalize(self)

    def update(self) -> None:
        if not self._preview_state.duration > 0:
            return

        self.render_static_graph(
            self._preview_state.position / self._preview_state.duration,
            self._seek_ts / self._preview_state.duration,
        )

    def _convert_position_to_timestamp(self, x: float, x_max: float) -> float:
        if x_max == 0.0:
            log.warning("Width is zero, when converting position to thimestamp")
            return 0.0

        if not self._preview_state.duration > 0:
            return 0.0

        if not self._is_ltr:
            x = x_max - x

        return self.convert_x_to_position(x) * self._preview_state.duration

    def _on_button_pressed(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        _y: float,
    ) -> None:
        self._user_holds_position_slider = True
        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        self._seek_ts = self._convert_position_to_timestamp(x, self.get_width())
        self.update()

    def _on_button_released(
        self,
        _gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        _y: float,
    ) -> None:
        if not self._preview_state.duration > 0:
            return

        timestamp = self._convert_position_to_timestamp(x, self.get_width())
        self._seek_ts = -1.0
        self.emit(
            "seeked",
            timestamp,
        )
        self._user_holds_position_slider = False
        self.render_static_graph(
            timestamp / self._preview_state.duration,
            self._seek_ts / self._preview_state.duration,
        )

    def _on_cursor_move(
        self,
        _event_controller: Gtk.EventControllerMotion,
        x: float,
        _y: float,
    ) -> None:
        if not self._user_holds_position_slider:
            return

        if abs(x - self._cursor_pos) < 1e-2:
            return

        self._cursor_pos = x
        self._seek_ts = self._convert_position_to_timestamp(x, self.get_width())
        self.update()

    def _on_scroll(
        self,
        _event_controller: Gtk.EventControllerScroll,
        dx: float,
        _dy: float,
    ) -> bool:
        if not self._preview_state.duration > 0:
            return False

        if dx > 0:
            timestamp = self._preview_state.position + self._scroll_step
        else:
            timestamp = self._preview_state.position - self._scroll_step
        self.emit("seeked", timestamp)
        self.render_static_graph(
            timestamp / self._preview_state.duration,
            self._seek_ts / self._preview_state.duration,
        )
        return True



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any
from typing import cast

import logging

from gi.repository import Adw
from gi.repository import Gtk
from nbxmpp.errors import StanzaError
from nbxmpp.protocol import JID
from nbxmpp.task import Task

from gajim.common import app

from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string

log = logging.getLogger("gajim.gtk.preferences.blocked_contacts")


@Gtk.Template.from_string(string=get_ui_string("preference/blocked_contacts.ui"))
class BlockedContacts(Gtk.Grid, SignalManager):
    __gtype_name__ = "BlockedContacts"

    _spinner: Adw.Spinner = Gtk.Template.Child()
    _block_view: Gtk.TreeView = Gtk.Template.Child()
    _jid_cell_renderer: Gtk.CellRendererText = Gtk.Template.Child()
    _add_button: Gtk.Button = Gtk.Template.Child()
    _remove_button: Gtk.Button = Gtk.Template.Child()
    _save_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self, account: str) -> None:
        Gtk.Grid.__init__(self)
        SignalManager.__init__(self)

        self._account = account
        self._client = app.get_client(account)
        self._prev_blocked_jids: set[JID] = set()

        self._blocking_store = Gtk.ListStore(str)
        self._block_view.set_model(self._blocking_store)

        self._set_grid_state(False)

        self._activate_spinner()

        self._connect(self._jid_cell_renderer, "edited", self._jid_edited)
        self._connect(self._add_button, "clicked", self._on_add)
        self._connect(self._remove_button, "clicked", self._on_remove)
        self._connect(self._save_button, "clicked", self._on_save)

        self._client.get_module("Blocking").request_blocking_list(
            callback=self._on_blocking_list_received
        )

    def _on_blocking_list_received(self, task: Task) -> None:
        self._disable_spinner()
        try:
            blocking_list = cast(set[JID], task.finish())
        except StanzaError as error:
            self._set_grid_state(False)
            log.error("Failed to retrieve blocked contacts: %s", error)
            return

        self._prev_blocked_jids = blocking_list
        self._blocking_store.clear()

        for jid in blocking_list:
            self._blocking_store.append([str(jid)])

        self._set_grid_state(True)

    def _on_save_result(self, task: Task):
        self._disable_spinner()
        try:
            task.finish()
        except StanzaError as error:
            log.error("Failed to store blocked contacts: %s", error)
        finally:
            self._set_grid_state(True)

    def _set_grid_state(self, state: bool) -> None:
        self.set_sensitive(state)

    def _jid_edited(
        self, _renderer: Gtk.CellRendererText, path: str, new_text: str
    ) -> None:
        iter_ = self._blocking_store.get_iter(path)
        self._blocking_store.set_value(iter_, 0, new_text)

    def _on_add(self, _button: Any) -> None:
        self._blocking_store.append([""])

    def _on_remove(self, _button: Any) -> None:
        selected_rows = self._block_view.get_selection().get_selected_rows()
        if selected_rows is None:
            return

        mod, paths = selected_rows
        for path in paths:
            iter_ = mod.get_iter(path)
            self._blocking_store.remove(iter_)

    def _on_save(self, _button: Gtk.Button) -> None:
        self._activate_spinner()
        self._set_grid_state(False)

        blocked_jids: set[JID] = set()
        for item in self._blocking_store:
            if not item[0]:
                # No address/placeholder
                continue
            blocked_jids.add(JID.from_string(item[0]))

        unblock_jids = self._prev_blocked_jids - blocked_jids
        block_jids = blocked_jids - self._prev_blocked_jids

        self._client.get_module("Blocking").update_blocking_list(
            block_jids,
            unblock_jids,
            callback=self._on_save_result,  # type: ignore
        )

    def _activate_spinner(self) -> None:
        self._spinner.set_visible(True)

    def _disable_spinner(self) -> None:
        self._spinner.set_visible(False)

    def do_unroot(self) -> None:
        self._disconnect_all()
        del self._client
        del self._blocking_store
        Gtk.Grid.do_unroot(self)
        app.check_finalize(self)
